/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/
#include "tileconfig.h"

#define SFLOAT4 uint2

cbuffer SImageSettings : register(b0)
{
	uint Nx;
	uint Ny;
	uint Tx;
	uint Ty;
	uint Spp;
	float DepthMin;
	float DepthMax;
	float Camera_C0;
	float Camera_C1;
	float MaxBlur;
	float sigma_x;
	float sigma_a;
	float sigma_t;
	float pad[3];
};

struct SLayerStats
{
	float smin;
	float smax;
	float vsumx;
	float vsumy;
	float v2sumx;
	float v2sumy;
	uint nSamples;
};

struct SLayerInfo
{
	int nSamples;
	float slopet;
	float sigmax;
	float sigmay;
	float sigmat;
	float sigmau;
	float sigmav;
	float shearxu;
	float shearyv;
	float angle;
};

float4 unpackSFloat4(SFLOAT4 d)
{
	float4 retval;

	retval.r = f16tof32(d.x & 0xffff);
	retval.g = f16tof32((d.x >> 16) & 0xffff);
	retval.b = f16tof32(d.y & 0xffff);
	retval.a = f16tof32((d.y >> 16) & 0xffff);

	return retval;
}

SFLOAT4 packSFloat4(float4 d)
{
	SFLOAT4 retval;

	retval.x = (f32tof16(d.g) << 16) | f32tof16(d.r);
	retval.y = (f32tof16(d.a) << 16) | f32tof16(d.b);
	return retval;
}

float ShearUV(float w) { return Camera_C0 / w + Camera_C1; }

float InvShearUV(float s)
{
	if (s > Camera_C1)
	{
		return GPU_FLT_MAX;
	}
	return Camera_C0 / (s - Camera_C1);
}

uint Getlayer(float z, float znorm)
{
	if (abs(Camera_C1) < GPU_EPS)
	{
		if (z > DepthMax) // Added a layer for DepthMax - inf to handle background stuff
		{ 
			return MAX_LAYERS - 1;
		}

		float zr = (z - DepthMin) * znorm;
		uint layer = min(MAX_LAYERS - 2, zr);
		return layer;
	}
	else
	{
		float zFocusStart = DOF_PART_MIN <= ShearUV(DepthMin) ? DepthMin : InvShearUV(DOF_PART_MIN);
		float zFocusEnd = DOF_PART_MAX >= ShearUV(DepthMax) ? DepthMax : InvShearUV(DOF_PART_MAX);

		float depthNear = max(1, zFocusStart - DepthMin);
		float depthFar = max(1, DepthMax - zFocusEnd);

		float layersNear = ceil((depthNear / (depthNear + depthFar)) * UNIFORM_LAYERS);
		float layersFar = UNIFORM_LAYERS - layersNear;

		if (z > DepthMax)
		{
			return COC_LAYERS + UNIFORM_LAYERS;
		}

		float s = ShearUV(z);
		if (z < zFocusStart)
		{
			return floor(layersNear * (z - DepthMin) / depthNear);
		}
		else if (z < zFocusEnd)
		{
			float step = ((DOF_PART_MAX - DOF_PART_MIN) - 3) / float(COC_LAYERS - 2);
			if (s < -1.5f)
			{
				return floor((s - (DOF_PART_MIN)) / step)
				       + layersNear; 
			}
			else if (s < 0.0f)
			{
				return (COC_LAYERS / 2 - 1) + layersNear;
			}
			else if (s < 1.5f)
			{
				return (COC_LAYERS / 2) + layersNear;
			}
			else
			{
				return floor((s - 1.5f) / step) + (COC_LAYERS / 2 + 1) + layersNear;
			}
		}
		else
		{
			return min(COC_LAYERS + UNIFORM_LAYERS - 1,
			           floor(layersFar * (z - zFocusEnd) / depthFar) + layersNear + COC_LAYERS);
		}
	}
}

float GetZNorm()
{
	float iznorm = max((DepthMax - DepthMin) / float(MAX_LAYERS - 1), GPU_DEPTH_EPS);
	return rcp(iznorm);
}

float4 UnquantizeXYUV(uint4 xyuv)
{
	float4 xyuvf;
    xyuvf.x = float(xyuv.x) / 255.0f;
    xyuvf.y = float(xyuv.y) / 255.0f;
    xyuvf.z = (float(xyuv.z) / 127.0f) - 1.0f;
    xyuvf.w = (float(xyuv.w) / 127.0f) - 1.0f;
    return xyuvf;
}

float UnquantizeT(uint t)
{

    float tf;
    tf = float(t) / 255.0f;
    return tf;
}

float ComputeAngle(float x, float y)
{
	float angle = 0;
	// Rotate such that the x-axis is aligned with the motion direction
	if (abs(y) < GPU_EPS)
	{
		angle = x > 0 ? 0.0f : GPU_PI;
	}
	else
	{
		angle = atan2(y, x);
	}

	return angle;
}
