/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/
#include "common.hlsl"

#define OUTER_STAT_RATIO (IMAGE_OUTER_TILE / STAT_TILE)

Texture2DArray<float4> BufC : register(t0);
Texture2DArray<float> BufZ : register(t1);
Texture2DArray<uint4> BufXYUV : register(t2);
Texture2DArray<uint> BufT : register(t3);
Texture2DArray<uint2> BufVel : register(t4);

RWStructuredBuffer<SLayerInfo> BufLayers : register(u0);
groupshared SLayerStats gs_layerStats[STAT_TILE * STAT_TILE];

SLayerInfo ComputeFilterData(SLayerStats gs_layerStats)
{
	SLayerInfo fi;
	fi.slopet = 0;
	fi.sigmax = 0;
	fi.sigmay = 0;
	fi.sigmat = 0;
	fi.sigmau = 0;
	fi.sigmav = 0;
	fi.shearxu = 0;
	fi.shearyv = 0;
	fi.angle = 0;
	fi.nSamples = gs_layerStats.nSamples;

	if (gs_layerStats.nSamples == 0)
	{
		return fi;
	}

	// Constants
	const float sigma_t_shutter = sigma_t;
	const float sigma_x_pix = sigma_x;
	const float sigma_u_lens = sigma_a;
	const float sigma_x_by_u = sigma_x_pix / sigma_u_lens;
	const float sigma_x_by_t = sigma_x_pix / sigma_t_shutter;
	const float FILTER_CLAMP = ceil((RING_SIZE + 0.5f) / 2.f);
	const float SHEAR_CLAMP = floor((RING_SIZE + 0.5f) / 2.f);

	float angle = 0;

	float frac = rcp(gs_layerStats.nSamples);
	float xsum = gs_layerStats.vsumx;
	float ysum = gs_layerStats.vsumy;
	float x2sum = gs_layerStats.v2sumx;
	float y2sum = gs_layerStats.v2sumy;

	// compute mean
	float xavg = frac * xsum;
	float yavg = frac * ysum;
	float std_avg_x = sqrt(abs(frac * x2sum - xavg * xavg));
	float std_avg_y = sqrt(abs(frac * y2sum - yavg * yavg));
	float radius = 1.0f * max(std_avg_x, std_avg_y);

	// Circle of confusion
	float s1 = gs_layerStats.smin;
	float s2 = gs_layerStats.smax;

	bool diffSignL = (s1 < 0) ^ (s2 < 0); // defocus
	float s1u = abs(s1);
	float s2u = abs(s2);
	float cmin = min(s1u, s2u);
	float cmax = max(s1u, s2u);

	// Rotate such that the x-axis is aligned with the motion direction
	angle = ComputeAngle(xavg, yavg);

	// Remap to rotated space
	float v_avg = sqrt(xavg * xavg + yavg * yavg);
	float vmin = max((float)GPU_EPS, v_avg - radius);
	float vmax = v_avg + radius;

	float sigmax = sigma_x_pix;
	float sigmay = sigma_x_pix;
	float sigmat = sigma_t_shutter;
	float sigmau = sigma_u_lens;
	float sigmav = sigma_u_lens;
	float slopet = 0.0f;
	float shearxu = 0;
	float shearyv = 0;

	// ======== BEGIN (V,Y) FILTER CODE ======================
	if (diffSignL && cmin >= sigma_x_by_u && cmax >= sigma_x_by_u)
	{
		sigmay = sigma_x_pix;
		sigmav = sigma_u_lens;
	}
	else if (cmin < sigma_x_by_u) // Check if this branch really helps
	{
		shearyv = (s1 + s2) / 2.f;
		sigmav = 2.0f * sigma_x_pix / (cmax - cmin + GPU_EPS); //  2*Omega_u^max/H
		sigmay = sigma_x_pix;
	}
	else
	{
		// Derive Egan's filter
		float c_avg = 2.0f * s1 * s2 / (s1 + s2);
		float tsigmax = 2.0f * sigma_u_lens * (cmax * cmin) / (cmax - cmin + GPU_EPS);
		float tsigmau = sigma_u_lens;
		tsigmax = min(FILTER_CLAMP, tsigmax); // Clamp sigmax to avoid truncation artifacts

		float a = tsigmax;
		float b = tsigmau;
		float c = c_avg;

		float as = a * a;
		float bs = b * b;
		float cs = c * c;

		float d = as + cs * bs;

		// more robust when a -> \infty
		float f = abs(a) < GPU_EPS ? 1 + bs : 1 + bs * cs / as;
		sigmay = max(sigma_x_pix, abs(b * c / sqrt(f)));
		sigmav = sqrt(d / cs);
		shearyv = c / f;
	}

	// ======== BEGIN (X, U,T) FILTER CODE ========================
	if (v_avg - radius < 0 && vmin >= sigma_x_by_t) // DoF Only
	{
		slopet = 0;
		sigmat = sigma_t_shutter;
		sigmau = sigmav;
		sigmax = sigmay;
		shearxu = shearyv;
		angle = 0;
	}
	else if (vmin < sigma_x_by_t && cmin < sigma_x_by_u) // Cone intersects Omega_x = 1/(2PI*sigma_x_pix) plane
	{
		slopet = (vmax + vmin) / 2.f;
		sigmat = 2.0f * sigma_x_pix / (vmax - vmin + GPU_EPS);
		sigmax = sigma_x_pix;
		sigmau = sigma_u_lens;
		shearxu = shearyv;
	}
	else if (vmin * sigma_t_shutter < cmin * sigma_u_lens) // Cone intersects Omega_u = 1/(2PI*sigma_u_lens) plane
	{
		sigmax = sigmay;
		sigmau = sigmav;
		shearxu = shearyv;

		slopet = (vmax + vmin) / 2.f;
		slopet = clamp(slopet, -SHEAR_CLAMP, SHEAR_CLAMP);
		float v_avg = abs(slopet);
		sigmat = sigma_u_lens * cmin / (vmax - v_avg + GPU_EPS);
		sigmat = max(sigmat, sigma_t_shutter);
	}
	else // Cone intersects Omega_t = sigma_t_shutter plane
	{
		v_avg = vmax > GPU_EPS ? 2.0f * vmin * vmax / (vmin + vmax) : 0;
		float tsigmat_x = 2.0f * sigma_t_shutter * (vmax * vmin) / (vmax - vmin + GPU_EPS);

		// Clamp sigmax to avoid truncation artifacts. Depends on reconstruction window
		tsigmat_x = min(FILTER_CLAMP, tsigmat_x);

		float a = tsigmat_x;
		float b = sigma_t_shutter;
		float c = v_avg;

		float a2 = a * a;
		float b2 = b * b;
		float c2 = c * c;

		float d = a2 + c2 * b2;

		// more robust formulation when a -> \infty
		float f = 1 + b2 * c2 / a2;
		sigmax = max(sigma_x_pix, abs(b * c / sqrt(f)));
		sigmat = sqrt(d / c2);
		slopet = c / f;

		shearxu = (s1 + s2) / 2.f;
		shearxu = clamp(shearxu, -SHEAR_CLAMP, SHEAR_CLAMP);
		float c_avg = abs(shearxu);
		sigmau = sigma_t_shutter * vmin / (cmax - c_avg + GPU_EPS);
		sigmau = max(sigmau, sigma_u_lens);
	}

	// Force x/y filter to a single pixel, can be used to check the result of the reprojection stage only
	//sigmax = sigma_x_pix;
	//sigmay = sigma_x_pix;
	//sigmat = sigma_t_shutter;
	//sigmau = sigma_u_lens;
	//sigmav = sigma_u_lens;

	fi.slopet = slopet;
	fi.sigmax = sigmax;
	fi.sigmay = sigmay;
	fi.sigmat = sigmat;
	fi.sigmau = sigmau;
	fi.sigmav = sigmav;
	fi.angle = angle;
	fi.shearxu = shearxu;
	fi.shearyv = shearyv;

	return fi;
}

void LayerReduce(uint tid_scalar, uint size)
{
	if (tid_scalar < size / 2)
	{
		uint idx = tid_scalar + size / 2;
		gs_layerStats[tid_scalar].vsumx += gs_layerStats[idx].vsumx;
		gs_layerStats[tid_scalar].vsumy += gs_layerStats[idx].vsumy;
		gs_layerStats[tid_scalar].v2sumx += gs_layerStats[idx].v2sumx;
		gs_layerStats[tid_scalar].v2sumy += gs_layerStats[idx].v2sumy;
		gs_layerStats[tid_scalar].nSamples += gs_layerStats[idx].nSamples;
		gs_layerStats[tid_scalar].smin = min(gs_layerStats[tid_scalar].smin, gs_layerStats[idx].smin);
		gs_layerStats[tid_scalar].smax = max(gs_layerStats[tid_scalar].smax, gs_layerStats[idx].smax);
	}
	GroupMemoryBarrierWithGroupSync();
}

void ComputeLayerStats(uint3 GTid, uint3 Gid)
{
	uint2 tile;
	tile.x = Gid.x;
	tile.y = Gid.y;

	SLayerStats layerStats[MAX_LAYERS];

	for (uint li = 0; li < MAX_LAYERS; ++li)
	{
		layerStats[li].vsumx = 0;
		layerStats[li].vsumy = 0;
		layerStats[li].v2sumx = 0;
		layerStats[li].v2sumy = 0;
		layerStats[li].nSamples = 0;
		layerStats[li].smin = GPU_FLT_MAX;
		layerStats[li].smax = -GPU_FLT_MAX;
	}

	int2 ref;
	ref.x = tile.x * IMAGE_INNER_TILE - RING_SIZE;
	ref.y = tile.y * IMAGE_INNER_TILE - RING_SIZE;

	float znorm = GetZNorm();

	uint tid_scalar = GTid.x + GTid.y * STAT_TILE;

	// Detect active bins
	for (uint i = 0; i < OUTER_STAT_RATIO; i++)
	{
		for (uint j = 0; j < OUTER_STAT_RATIO; j++)
		{
			uint x = GTid.x + j * STAT_TILE;
			uint y = GTid.y + i * STAT_TILE;

			int3 coord;
			coord.x = ref.x + x;
			coord.y = ref.y + y;

			if (coord.x >= 0 && coord.y >= 0 && coord.x < int(Nx) && coord.y < int(Ny))
			{
				for (uint k = 0; k < Spp; k++)
				{
					coord.z = k;

					// Find layer based on depth
					float z = BufZ[coord];
					uint layer = Getlayer(z, znorm);

					float shear = ShearUV(z);
					shear = max(-MaxBlur, min(MaxBlur, shear));
					layerStats[layer].smin = min(layerStats[layer].smin, shear);
					layerStats[layer].smax = max(layerStats[layer].smax, shear);

					// Fetch sample velocity
					float2 vel = (float2(BufVel[coord]) - 127.5f) * MaxBlur / 127.5f;
					float2 vel2 = vel * vel;

					layerStats[layer].vsumx += vel.x;
					layerStats[layer].vsumy += vel.y;
					layerStats[layer].v2sumx += vel2.x;
					layerStats[layer].v2sumy += vel2.y;
					layerStats[layer].nSamples += 1;
				}
			}
		}
	}

	// Second level reductions
	for (i = 0; i < MAX_LAYERS; ++i)
	{
		// Copy to Shared mem
		gs_layerStats[tid_scalar].vsumx = layerStats[i].vsumx;
		gs_layerStats[tid_scalar].vsumy = layerStats[i].vsumy;
		gs_layerStats[tid_scalar].v2sumx = layerStats[i].v2sumx;
		gs_layerStats[tid_scalar].v2sumy = layerStats[i].v2sumy;
		gs_layerStats[tid_scalar].smin = layerStats[i].smin;
		gs_layerStats[tid_scalar].smax = layerStats[i].smax;
		gs_layerStats[tid_scalar].nSamples = layerStats[i].nSamples;
		GroupMemoryBarrierWithGroupSync();

		LayerReduce(tid_scalar, 64);
		LayerReduce(tid_scalar, 32);
		LayerReduce(tid_scalar, 16);
		LayerReduce(tid_scalar, 8);
		LayerReduce(tid_scalar, 4);
		LayerReduce(tid_scalar, 2);

		if (tid_scalar == 0)
		{
			int tileIdx = (tile.x + tile.y * Tx) * MAX_LAYERS;
			BufLayers[tileIdx + i] = ComputeFilterData(gs_layerStats[0]);
		}
	}
}

[numthreads(STAT_TILE, STAT_TILE, 1)] void CSMain(uint3 GTid : SV_GroupThreadID, uint3 Gid : SV_GroupID)
{
	ComputeLayerStats(GTid, Gid);
}
