/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/
#include "common.hlsl"

Texture2DArray<float4> BufC : register(t0);
Texture2DArray<float> BufZ : register(t1);
Texture2DArray<uint4> BufXYUV : register(t2);
Texture2DArray<uint> BufT : register(t3);
Texture2DArray<uint2> BufVel : register(t4);
StructuredBuffer<SLayerInfo> BufLayers : register(t5);

RWStructuredBuffer<SFLOAT4> BufComposite : register(u0);
groupshared int4 gs_Col[IMAGE_OUTER_TILE][IMAGE_OUTER_TILE];

SLayerInfo ComputeFilterData(uint3 GTid, uint3 Gid, uint layer)
{
	int2 tile;
	tile.x = Gid.x;
	tile.y = Gid.y;

	return BufLayers[(tile.x + tile.y * Tx) * MAX_LAYERS + layer];
}

void ResetTile(uint3 GTid, uint3 Gid)
{
	for (int i = 0; i < OUTER_COMPUTE_RATIO; i++)
	{
		for (int j = 0; j < OUTER_COMPUTE_RATIO; j++)
		{
			// Reset pre-integration buffer
			uint tilex = GTid.x + j * IMAGE_INNER_TILE;
			uint tiley = GTid.y + i * IMAGE_INNER_TILE;

			if (tilex < IMAGE_OUTER_TILE && tiley < IMAGE_OUTER_TILE)
			{
				gs_Col[tilex][tiley] = 0;
			}
		}
	}

	GroupMemoryBarrierWithGroupSync();
}

#ifndef REPROJECT_BILINEAR
// Nearest
void SplatColor(float x, float y, float4 rgba, float w)
{
	int xi = (int)floor(x);
	int yi = (int)floor(y);

	uint wquant = min(w * WU_SCALE, WU_SCALE);
	uint4 rgbau
	    = min((int4)(rgba * COLOR_SCALE), COLOR_CLAMP); // Convert to 16 bit int. Has to be clamped to some limit.
	rgbau *= wquant;                                    // Fits in 24 bits

	InterlockedAdd(gs_Col[yi][xi].r, rgbau.r);
	InterlockedAdd(gs_Col[yi][xi].g, rgbau.g);
	InterlockedAdd(gs_Col[yi][xi].b, rgbau.b);

	uint opacity = (wquant << WU_SHIFT) | wquant;
	InterlockedAdd(gs_Col[yi][xi].a, opacity);
}
void SplatWeight(float x, float y, float w)
{
	int xi = (int)floor(x);
	int yi = (int)floor(y);

	uint wquant = min(w * WU_SCALE, WU_SCALE);
	InterlockedAdd(gs_Col[yi][xi].a, wquant);
}
#else
// Bilerp
void SplatColor(float x, float y, float4 rgba, float w)
{
	int x1 = clamp((int)floor(x - 0.5f), 0, IMAGE_OUTER_TILE - 1);
	int y1 = clamp((int)floor(y - 0.5f), 0, IMAGE_OUTER_TILE - 1);
	int x2 = clamp((int)floor(x - 0.5f) + 1, 0, IMAGE_OUTER_TILE - 1);
	int y2 = clamp((int)floor(y - 0.5f) + 1, 0, IMAGE_OUTER_TILE - 1);

	float f_x = frac(x);
	float f_y = frac(y);

	float w00 = (1.0f - f_y) * (1.0f - f_x);
	float w10 = (f_y) * (1.0f - f_x);
	float w01 = (1.0f - f_y) * (f_x);
	float w11 = (f_y) * (f_x);

	uint wu00 = min(w * w00 * WU_SCALE, WU_SCALE);
	uint wu01 = min(w * w01 * WU_SCALE, WU_SCALE);
	uint wu10 = min(w * w10 * WU_SCALE, WU_SCALE);
	uint wu11 = min(w * w11 * WU_SCALE, WU_SCALE);

	uint4 rgbau00 = (uint4)min(rgba * COLOR_SCALE, COLOR_CLAMP) * wu00;
	uint4 rgbau01 = (uint4)min(rgba * COLOR_SCALE, COLOR_CLAMP) * wu01;
	uint4 rgbau10 = (uint4)min(rgba * COLOR_SCALE, COLOR_CLAMP) * wu10;
	uint4 rgbau11 = (uint4)min(rgba * COLOR_SCALE, COLOR_CLAMP) * wu11;

	InterlockedAdd(gs_Col[y1][x1].r, rgbau00.r);
	InterlockedAdd(gs_Col[y1][x1].g, rgbau00.g);
	InterlockedAdd(gs_Col[y1][x1].b, rgbau00.b);
	InterlockedAdd(gs_Col[y1][x1].a, (wu00 << WU_SHIFT) | wu00);

	InterlockedAdd(gs_Col[y1][x2].r, rgbau01.r);
	InterlockedAdd(gs_Col[y1][x2].g, rgbau01.g);
	InterlockedAdd(gs_Col[y1][x2].b, rgbau01.b);
	InterlockedAdd(gs_Col[y1][x2].a, (wu01 << WU_SHIFT) | wu01);

	InterlockedAdd(gs_Col[y2][x1].r, rgbau10.r);
	InterlockedAdd(gs_Col[y2][x1].g, rgbau10.g);
	InterlockedAdd(gs_Col[y2][x1].b, rgbau10.b);
	InterlockedAdd(gs_Col[y2][x1].a, (wu10 << WU_SHIFT) | wu10);

	InterlockedAdd(gs_Col[y2][x2].r, rgbau11.r);
	InterlockedAdd(gs_Col[y2][x2].g, rgbau11.g);
	InterlockedAdd(gs_Col[y2][x2].b, rgbau11.b);
	InterlockedAdd(gs_Col[y2][x2].a, (wu11 << WU_SHIFT) | wu11);
}
void SplatWeight(float x, float y, float w)
{
	int x1 = clamp((int)floor(x), 0, IMAGE_OUTER_TILE - 1);
	int y1 = clamp((int)floor(y), 0, IMAGE_OUTER_TILE - 1);
	int x2 = clamp((int)floor(x) + 1, 0, IMAGE_OUTER_TILE - 1);
	int y2 = clamp((int)floor(y) + 1, 0, IMAGE_OUTER_TILE - 1);

	float f_x = frac(x);
	float f_y = frac(y);

	float w00 = (1.0f - f_y) * (1.0f - f_x);
	float w10 = (f_y) * (1.0f - f_x);
	float w01 = (1.0f - f_y) * (f_x);
	float w11 = (f_y) * (f_x);

	uint wu00 = min(w * w00 * WU_SCALE, WU_SCALE);
	uint wu01 = min(w * w01 * WU_SCALE, WU_SCALE);
	uint wu10 = min(w * w10 * WU_SCALE, WU_SCALE);
	uint wu11 = min(w * w11 * WU_SCALE, WU_SCALE);

	InterlockedAdd(gs_Col[y1][x1].a, wu00);
	InterlockedAdd(gs_Col[y1][x2].a, wu01);
	InterlockedAdd(gs_Col[y2][x1].a, wu10);
	InterlockedAdd(gs_Col[y2][x2].a, wu11);
}
#endif

void Reproject(uint3 GTid, uint3 Gid, uint currLayer, SLayerInfo fi)
{
	// Outer tile coordinates
	int x0 = Gid.x * IMAGE_INNER_TILE - RING_SIZE;
	int y0 = Gid.y * IMAGE_INNER_TILE - RING_SIZE;

	float znorm = GetZNorm();

	float s_a = sin(fi.angle);
	float c_a = cos(fi.angle);

	// Rotate motion vector:  R [v_x, 0]^T
	float vx_r = c_a * fi.slopet;
	float vy_r = s_a * fi.slopet;

	// Rotate defocus parameters:  R A R^T, where A is diagonal
	float a00 = c_a * c_a * fi.shearxu + s_a * s_a * fi.shearyv;
	float a01 = s_a * c_a * (fi.shearxu - fi.shearyv);
	float a10 = a01;
	float a11 = c_a * c_a * fi.shearyv + s_a * s_a * fi.shearxu;

	float uweight = -0.5f * rcp(fi.sigmau * fi.sigmau);
	float vweight = -0.5f * rcp(fi.sigmav * fi.sigmav);
	float tweight = -0.5f * rcp(fi.sigmat * fi.sigmat);

	for (int i = 0; i < OUTER_COMPUTE_RATIO; i++)
	{
		for (int j = 0; j < OUTER_COMPUTE_RATIO; j++)
		{
			uint tilex = GTid.x + j * IMAGE_INNER_TILE;
			uint tiley = GTid.y + i * IMAGE_INNER_TILE;

			int3 coord;
			coord.x = x0 + tilex;
			coord.y = y0 + tiley;

			// Re-Project samples
			if (tilex < IMAGE_OUTER_TILE && tiley < IMAGE_OUTER_TILE 
				&& coord.x >= 0 && coord.y >= 0
			    && coord.x < int(Nx) && coord.y < int(Ny))
			{
				for (uint k = 0; k < Spp; k++)
				{
					coord.z = k;

					// Find layer based on depth
					float z = BufZ[coord];
					uint layer = Getlayer(z, znorm);

					if (layer >= currLayer)
					{
						uint4 xyuv = BufXYUV[coord];
						float4 xyuvf = UnquantizeXYUV(xyuv);
						xyuvf.x += tilex;
						xyuvf.y += tiley;

						uint t = BufT[coord];
						float tf = UnquantizeT(t) - 0.5f;

						// Reproject to the center of the lens and t=0 and accumulate
						float x0 = xyuvf.x - a00 * xyuvf.z - a01 * xyuvf.w - vx_r * tf;
						float y0 = xyuvf.y + a10 * xyuvf.z + a11 * xyuvf.w + vy_r * tf; 

						// Check Bounds
						if (x0 >= -0.5 && x0 < (float)IMAGE_OUTER_TILE-0.5f && y0 >= -0.5f && y0 < (float)IMAGE_OUTER_TILE-0.5f)
						{
							float w = exp(uweight * (xyuvf.z * xyuvf.z) + vweight * (xyuvf.w * xyuvf.w)
							              + tweight * (tf * tf));

							if (layer == currLayer) // Matched partition
							{
								SplatColor(x0, y0, BufC[coord], w);
							}
							else
							{
								SplatWeight(x0, y0, w);
							}
						}
					}
				}
			}
		}
	}

	GroupMemoryBarrierWithGroupSync();
}

#ifndef FILTER_BILINEAR
// Nearest
inline float3 filter_fixed(float x, float y, out float Nr, out float Nl)
{
	int xi = clamp((int)floor(x + 0.5f), 0, IMAGE_OUTER_TILE - 1);
	int yi = clamp((int)floor(y + 0.5f), 0, IMAGE_OUTER_TILE - 1);
	Nr = (float)(gs_Col[yi][xi].a & WU_MASK);
	Nl = (float)((gs_Col[yi][xi].a >> WU_SHIFT) & WU_MASK);
	return (gs_Col[yi][xi].rgb);
}
#else
// Bilerp
inline float3 filter_fixed(float x, float y, out float Nr, out float Nl)
{
	int x1 = clamp((int)floor(x), 0, IMAGE_OUTER_TILE - 1);
	int y1 = clamp((int)floor(y), 0, IMAGE_OUTER_TILE - 1);
	int x2 = clamp((int)floor(x)+1, 0, IMAGE_OUTER_TILE - 1);
	int y2 = clamp((int)floor(y)+1, 0, IMAGE_OUTER_TILE - 1);

	int4 c00 = gs_Col[y1][x1];
	int4 c10 = gs_Col[y2][x1];
	int4 c01 = gs_Col[y1][x2];
	int4 c11 = gs_Col[y2][x2];

	float f_x = frac(x);
	float f_y = frac(y);

	float w00 = (1.0f - f_y) * (1.0f - f_x);
	float w10 = (f_y) * (1.0f - f_x);
	float w01 = (1.0f - f_y) * (f_x);
	float w11 = (f_y) * (f_x);

	Nr =  w00 * (c00.a & WU_MASK) + w10 * (c10.a & WU_MASK) 
		+ w01 * (c01.a & WU_MASK) + w11 * (c11.a & WU_MASK);
	Nl =  w00 * ((c00.a >> WU_SHIFT) & WU_MASK) + w10 * ((c10.a >> WU_SHIFT) & WU_MASK) 
		+ w01 * ((c01.a >> WU_SHIFT) & WU_MASK) + w11 * ((c11.a >> WU_SHIFT) & WU_MASK);

	return w00 * c00.rgb + w10 * c10.rgb + w01 * c01.rgb + w11 * c11.rgb;
}
#endif

void FilterX(uint3 GTid, uint3 Gid, SLayerInfo fi)
{
	// Thread Limits
	int x0 = GTid.x;
	int y0 = GTid.y;

	// Get Filter Width
	float scale = rcp(fi.sigmax * fi.sigmax);
	int radius = min(int(fi.sigmax * 4.0f), RING_SIZE);
	int yradius = min(int(fi.sigmay * 4.0f), RING_SIZE);
	float angle = fi.angle;

	float c_a = cos(angle);
	float s_a = sin(angle);

	float color[OUTER_COMPUTE_RATIO * OUTER_COMPUTE_RATIO][3];
	float wt[OUTER_COMPUTE_RATIO * OUTER_COMPUTE_RATIO][2];

	// Compute guard band
	float xstep = abs(s_a * yradius);
	float ystep = abs(c_a * yradius);

	uint xstart = floor(RING_SIZE - xstep);
	uint xend = ceil(IMAGE_INNER_TILE + RING_SIZE + xstep);
	uint ystart = floor(RING_SIZE - ystep);
	uint yend = ceil(IMAGE_INNER_TILE + RING_SIZE + ystep);

	for (uint y = 0; y < OUTER_COMPUTE_RATIO; y++)
	{
		for (uint x = 0; x < OUTER_COMPUTE_RATIO; x++)
		{
			uint x1 = x * IMAGE_INNER_TILE + x0;
			uint y1 = y * IMAGE_INNER_TILE + y0;

			if (x1 >= xstart && x1 <= xend && y1 >= ystart && y1 <= yend)
			{
				float3 col = 0;
				wt[x + y * OUTER_COMPUTE_RATIO][0] = 0;
				wt[x + y * OUTER_COMPUTE_RATIO][1] = 0;

				for (int k = -radius; k <= radius; k++)
				{
					float x2 = x1 + k * c_a;
					float y2 = y1 - k * s_a;

					float w = exp(-0.5f * k * k * scale);
					float Nr, Nl;
					float3 c = filter_fixed(x2, y2, Nr, Nl);

					col += w * c;
					wt[x + y * OUTER_COMPUTE_RATIO][0] += w * Nr;
					wt[x + y * OUTER_COMPUTE_RATIO][1] += w * Nl;
				}

				color[x + y * OUTER_COMPUTE_RATIO][0] = col.x;
				color[x + y * OUTER_COMPUTE_RATIO][1] = col.y;
				color[x + y * OUTER_COMPUTE_RATIO][2] = col.z;
			}
		}
	}

	GroupMemoryBarrierWithGroupSync();

	for (uint yy = 0; yy < OUTER_COMPUTE_RATIO; yy++)
	{
		for (uint x = 0; x < OUTER_COMPUTE_RATIO; x++)
		{
			uint x1 = x * IMAGE_INNER_TILE + x0;
			uint y1 = yy * IMAGE_INNER_TILE + y0;

			if (x1 < IMAGE_OUTER_TILE && y1 < IMAGE_OUTER_TILE)
			{
				gs_Col[y1][x1].r = asuint(color[x + yy * OUTER_COMPUTE_RATIO][0]);
				gs_Col[y1][x1].g = asuint(color[x + yy * OUTER_COMPUTE_RATIO][1]);
				gs_Col[y1][x1].b = asuint(color[x + yy * OUTER_COMPUTE_RATIO][2]);

				uint Nru = f32tof16(wt[x + yy * OUTER_COMPUTE_RATIO][0]);
				uint Nlu = f32tof16(wt[x + yy * OUTER_COMPUTE_RATIO][1]);

				gs_Col[y1][x1].a = (Nlu << 16) | Nru;
			}
		}
	}

	GroupMemoryBarrierWithGroupSync();
}

void Composite(uint3 GTid, uint3 Gid, float3 color, float opacity[2])
{
	uint x0 = Gid.x * IMAGE_INNER_TILE + GTid.x;
	uint y0 = Gid.y * IMAGE_INNER_TILE + GTid.y;

	if (x0 < Nx && y0 < Ny)
	{
		uint coord = y0 * Nx + x0;
		float4 colori = unpackSFloat4(BufComposite[coord]);
		float norm = rcp(opacity[0] + GPU_EPS);
		float alpha = opacity[1] * norm;
		float scale = colori.a * norm;

		float4 coloro;
		coloro.rgb = colori.rgb + scale * color;
		coloro.a = colori.a * (1 - alpha);

		BufComposite[coord] = packSFloat4(coloro);
	}

	GroupMemoryBarrierWithGroupSync();
}

#ifndef FILTER_BILINEAR
// Nearest
inline float3 filter_cast(float x, float y, out float Nr, out float Nl)
{
	int xi = clamp((int)floor(x + 0.5f), 0, IMAGE_OUTER_TILE - 1);
	int yi = clamp((int)floor(y + 0.5f), 0, IMAGE_OUTER_TILE - 1);

	Nr = f16tof32(gs_Col[yi][xi].a & 0xffff);
	Nl = f16tof32((gs_Col[yi][xi].a >> 16) & 0xffff);
	return asfloat(gs_Col[yi][xi].rgb);
}
#else
// Bilerp
inline float3 filter_cast(float x, float y, out float Nr, out float Nl)
{
	int x1 = clamp((int)floor(x), 0, IMAGE_OUTER_TILE - 1);
	int y1 = clamp((int)floor(y), 0, IMAGE_OUTER_TILE - 1);
	int x2 = clamp((int)floor(x)+1, 0, IMAGE_OUTER_TILE - 1);
	int y2 = clamp((int)floor(y)+1, 0, IMAGE_OUTER_TILE - 1);

	int4 c00 = gs_Col[y1][x1];
	int4 c10 = gs_Col[y2][x1];
	int4 c01 = gs_Col[y1][x2];
	int4 c11 = gs_Col[y2][x2];

	float f_x = frac(x);
	float f_y = frac(y);

	float w00 = (1.0f - f_y) * (1.0f - f_x);
	float w10 = (f_y) * (1.0f - f_x);
	float w01 = (1.0f - f_y) * (f_x);
	float w11 = (f_y) * (f_x);

	Nr =   w00 * f16tof32(c00.a & 0xffff) + w10 * f16tof32(c10.a & 0xffff) 
		 + w01 * f16tof32(c01.a & 0xffff) + w11 * f16tof32(c11.a & 0xffff);
	Nl =   w00 * f16tof32((c00.a >> 16) & 0xffff) + w10 * f16tof32((c10.a >> 16) & 0xffff)
	     + w01 * f16tof32((c01.a >> 16) & 0xffff) + w11 * f16tof32((c11.a >> 16) & 0xffff);
	return w00 * asfloat(c00.rgb) + w10 * asfloat(c10.rgb) + w01 * asfloat(c01.rgb) + w11 * asfloat(c11.rgb);
}
#endif

void FilterY(uint3 GTid, uint3 Gid, SLayerInfo fi)
{
	// Thread Limits
	int x0 = GTid.x + RING_SIZE;
	int y0 = GTid.y + RING_SIZE;

	// Get Filter Width
	float scale = rcp(fi.sigmay * fi.sigmay);
	int radius = min(int(fi.sigmay * 4.0f), RING_SIZE);
	float angle = fi.angle;

	float c_a = cos(angle);
	float s_a = sin(angle);

	float3 color = 0;
	float wt[2] = {0, 0};

	for (int k = -radius; k <= radius; k++)
	{
		float x2 = x0 + k * s_a;
		float y2 = y0 + k * c_a;

		float w = exp(-0.5f * k * k * scale);

		float Nr, Nl;
		float3 c = filter_cast(x2, y2, Nr, Nl);

		color += w * c;
		wt[0] += w * Nr;
		wt[1] += w * Nl;
	}

	GroupMemoryBarrierWithGroupSync();

	// Composite
	Composite(GTid, Gid, color, wt);
}

void ResetComposite(uint3 GTid, uint3 Gid)
{
	uint x0 = Gid.x * IMAGE_INNER_TILE + GTid.x;
	uint y0 = Gid.y * IMAGE_INNER_TILE + GTid.y;

	if (x0 < Nx && y0 < Ny)
	{
		uint coord = y0 * Nx + x0;
		float4 coloro;
		coloro.r = 0;
		coloro.g = 0;
		coloro.b = 0;
		coloro.a = 1;

		BufComposite[coord] = packSFloat4(coloro);
	}
}

[numthreads(IMAGE_INNER_TILE, IMAGE_INNER_TILE, 1)] void CSMain(uint3 GTid : SV_GroupThreadID, uint3 Gid : SV_GroupID)
{
	ResetComposite(GTid, Gid);
	for (uint i = 0; i < MAX_LAYERS; i++)
	{
		SLayerInfo fi = ComputeFilterData(GTid, Gid, i);
		if (fi.nSamples > 0)
		{
			ResetTile(GTid, Gid);
			Reproject(GTid, Gid, i, fi);
			FilterX(GTid, Gid, fi);
			FilterY(GTid, Gid, fi);
		}
	}
}
