#pragma once

/*
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include <immintrin.h>
#include <assert.h>
#include <math.h>

#define EPS 1E-8f
#define SCENE_MAX_DEPTH 100000.0f

#define SAFE_RELEASE(p)     \
	{                       \
		if (p)              \
		{                   \
			(p)->Release(); \
			(p) = NULL;     \
		}                   \
	}
#define SAFE_DELETE(p)  \
	{                   \
		if (p)          \
		{               \
			delete (p); \
			(p) = NULL; \
		}               \
	}
#define SAFE_DELETE_ARRAY(p) \
	{                        \
		if ((p) != NULL)     \
		{                    \
			delete[](p);     \
			(p) = NULL;      \
		}                    \
	}

#ifdef M_PI
#undef M_PI
#endif
#define M_PI 3.14159265358979323846f

template <class T> T clamp(T in, T low, T high)
{
	if (in <= low)
	{
		return low;
	}
	else if (in >= high)
	{
		return high;
	}
	else
	{
		return in;
	}
}

struct SampleInfoHeader
{
	unsigned int xResolution;
	unsigned int yResolution;
	unsigned int spp;
	float C0; // DOF parameters (thin lens model)
	float C1;
};

struct SampleInfo
{
	float rgb[3];
	float imageX;
	float imageY;
	float lensU;
	float lensV;
	float time;
	float w;
	float vx, vy, vz;

	SampleInfo() {}
};

struct float2
{
	float x, y;

	float2() : x(0), y(0) {}
	float2(float xx, float yy)
	{
		x = xx;
		y = yy;
	}

	float2 operator*(float f) const { return float2(x * f, y * f); }
	float2 operator/(float f) const { return float2(x / f, y / f); }
	float2 operator-(const float2 &f2) const { return float2(x - f2.x, y - f2.y); }
	float2 operator+(const float2 &f2) const { return float2(x + f2.x, y + f2.y); }
	float2 operator*(const float2 &f2) const { return float2(x * f2.x, y * f2.y); }

	float2 &operator+=(const float2 &f2)
	{
		x += f2.x;
		y += f2.y;
		return *this;
	}

	float2 &operator*=(const float &f)
	{
		x *= f;
		y *= f;
		return *this;
	}
};

struct float3
{
	float x, y, z;

	float3() : x(0), y(0), z(0) {}
	float3(float xx, float yy, float zz)
	{
		x = xx;
		y = yy;
		z = zz;
	}

	float3 operator*(float f) const { return float3(x * f, y * f, z * f); }
	float3 operator/(float f) const { return float3(x / f, y / f, z / f); }
	float3 operator-(const float3 &f2) const { return float3(x - f2.x, y - f2.y, z - f2.z); }
	float3 operator+(const float3 &f2) const { return float3(x + f2.x, y + f2.y, z + f2.z); }
	float3 operator*(const float3 &f2) const { return float3(x * f2.x, y * f2.y, z * f2.z); }

	float3 &operator+=(const float3 &f2)
	{
		x += f2.x;
		y += f2.y;
		z += f2.z;
		return *this;
	}
};

// Alignment
#define SSE_ALIGN __declspec(align(16))

SSE_ALIGN struct float4
{
	__m128 m;

	float4() { m = _mm_set1_ps(0); }

	float4(float val) { m = _mm_set1_ps(val); }

	float4(__m128 v) : m(v) {}

	float4(float x, float y, float z, float w) { m = _mm_set_ps(w, z, y, x); }

	float4(float3 rgb, float w) { m = _mm_set_ps(w, rgb.z, rgb.y, rgb.x); }

	float get(unsigned int i) const
	{
		assert(i < 4);
		return m.m128_f32[i];
	}
	void set(unsigned int i, float v)
	{
		assert(i < 4);
		m.m128_f32[i] = v;
	}

	float4 operator+(const float4 &rhs) const { return float4(_mm_add_ps(m, rhs.m)); }
	float4 operator-(const float4 &rhs) const { return float4(_mm_sub_ps(m, rhs.m)); }
	float4 operator*(const float4 &rhs) const { return float4(_mm_mul_ps(m, rhs.m)); }
	float4 operator/(const float4 &rhs) const { return float4(_mm_div_ps(m, rhs.m)); }

	float4 &operator+=(const float4 &rhs)
	{
		m = _mm_add_ps(m, rhs.m);
		return *this;
	}
	float4 &operator-=(const float4 &rhs)
	{
		m = _mm_sub_ps(m, rhs.m);
		return *this;
	}
	float4 &operator*=(const float4 &rhs)
	{
		m = _mm_mul_ps(m, rhs.m);
		return *this;
	}
	float4 &operator/=(const float4 &rhs)
	{
		m = _mm_div_ps(m, rhs.m);
		return *this;
	}

	friend float4 operator*(const float lhs, const float4 &rhs) { return float4(lhs) * rhs; }
	friend float4 saturate(const float4 &v)
	{
		return float4(_mm_min_ps(_mm_max_ps(v.m, _mm_setzero_ps()), _mm_set1_ps(1.f)));
	}


	friend float4 saturate_low(const float4& v) { return float4( _mm_max_ps(v.m, _mm_setzero_ps()) ); }

	friend float4 pow(const float4 &lhs, const float exponent)
	{
		float4 ret;
		for (int i = 0; i < 4; ++i) {
			ret.m.m128_f32[i] = powf(lhs.m.m128_f32[i], exponent);
		}
		return ret;
	}
};
