/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include <cstring>
#include <stdio.h>
#include <stdlib.h>

#include "config.h"

/////////////////////////////////////////////////////////////////////////////////
// Globals
/////////////////////////////////////////////////////////////////////////////////

const StrEnumTuple Config::ALGO_LIST[NUM_ALGOS]
    = {StrEnumTuple(RECON_GAUSS, "gauss"), StrEnumTuple(RECON_5D, "5d"), StrEnumTuple(RECON_5D_GPU, "5d_gpu")};

Config *pConfig;

/////////////////////////////////////////////////////////////////////////////////
// Configuration
/////////////////////////////////////////////////////////////////////////////////

const char *boolToStr(bool b) { return b ? "true" : "false"; }

Config::Config()
{
	DepthLayers = 16;
	RecWin = 32;
	BlockSize = 16;
	DumpLayer = 0;
	MaxMotion = 32.0f;
	sigma_x = 1.0f / 3.0f;
	sigma_a = 1.0f / 3.0f;
	sigma_t = 1.0f / 3.0f;
	Algo = RECON_5D;
	LookupFilter = FILTER_BILINEAR;
}

void Config::PrintUsage()
{
	printf("Usage: reconstruct.exe [options] <trace_name.bin>\n");
	printf("\nOptions:\n");
	printf("    -algo [gauss | 5d | 5d_gpu]	Selects reconstruction algorithm [default=5d]\n");
	printf("    -maxblur <size>             Sets the maximum blur (larger will be clamped) [default=32]\n");
	printf("    -sigma_x <number>           Sets the pixel filter width [default=0.33333]\n");
	printf("    -sigma_a <number>           Sets the aperture filter width [default=0.33333]\n");
	printf("    -sigma_t <number>           Sets the shutter filter width [default=0.33333]\n");
	printf("\nOptions for -algo 5d:\n");
	printf("    -blocksize <size>           Sets the tile size  [default=16]\n");
	printf("    -guardsize <size>           Sets the size of the guard band [default=32]\n");
	printf("    -depthlayers <number>       Sets the minimum (uniform) layers [default=16]\n");
	printf("    -filter [nearest|linear]    Sets the filtering mode used for buffer access [default=linear]\n");
	printf("\nOptions for -algo 5d_gpu:\n");
	printf("    The gpu algorithm can only be configured using the defines in tileconfig.h in the source code\n");
	exit(1);
}

void Config::Print()
{
	printf("Config:\n");
	printf("    Algorithm: %s\n", ALGO_LIST[Algo].Name);
	printf("    MaxBlur: %2.2f\n", MaxMotion);
	printf("    sigma_x: %2.2f\n", sigma_x);
	printf("    sigma_a: %2.2f\n", sigma_a);
	printf("    sigma_t: %2.2f\n", sigma_t);
	if (Algo == RECON_5D)
	{
		printf("    BlockSize: %d\n", BlockSize);
		printf("    GuardSize: %d\n", RecWin);
		printf("    DepthLayers: %d\n", DepthLayers);
	}
	printf("-------------------------------\n");
}

void Config::Parse(int argc, const char *argv[])
{
	for (int arg = 1; arg < argc - 1; ++arg)
	{
		if (_stricmp(argv[arg], "-algo") == 0)
		{
			if (_stricmp(argv[arg + 1], "gauss") == 0)
			{
				Algo = RECON_GAUSS;
			}
			else if (_stricmp(argv[arg + 1], "5d") == 0)
			{
				Algo = RECON_5D;
			}
			else if (_stricmp(argv[arg + 1], "5d_gpu") == 0)
			{
				Algo = RECON_5D_GPU;
			}
			else
			{
				PrintUsage();
			}

			arg++;
		}
		else if (_stricmp(argv[arg], "-filter") == 0) 
		{
			if (_stricmp(argv[arg + 1], "nearest") == 0)
			{
				LookupFilter = FILTER_NEAREST;
			}
			else if (_stricmp(argv[arg + 1], "linear") == 0)
			{
				LookupFilter = FILTER_BILINEAR;
			}
			else
			{
				PrintUsage();
			}
			
			arg++;
		}
		else if (_stricmp(argv[arg], "-blocksize") == 0)
		{
			BlockSize = atoi(argv[arg + 1]);
			arg++;

			if (BlockSize < 4 || BlockSize > 64)
			{
				PrintUsage();
			}
		}
		else if (_stricmp(argv[arg], "-guardsize") == 0)
		{
			RecWin = atoi(argv[arg + 1]);
			arg++;

			if (RecWin < 4 || RecWin > 128 || RecWin < BlockSize)
			{
				PrintUsage();
			}
		}
		else if (_stricmp(argv[arg], "-maxblur") == 0)
		{
			MaxMotion = (float)atof(argv[arg + 1]);
			arg++;

			if (MaxMotion < 1 || MaxMotion > 1024)
			{
				PrintUsage();
			}
		}
		else if (_stricmp(argv[arg], "-depthlayers") == 0)
		{
			DepthLayers = atoi(argv[arg + 1]);
			arg++;

			if (DepthLayers < 0 || DepthLayers > 64)
			{
				PrintUsage();
			}
		}
		else if (_stricmp(argv[arg], "-sigma_x") == 0)
		{
			sigma_x = (float)atof(argv[arg + 1]);
			arg++;

			if (sigma_x < 0 || sigma_x > 10.0f)
			{
				PrintUsage();
			}
		}
		else if (_stricmp(argv[arg], "-sigma_a") == 0)
		{
			sigma_a = (float)atof(argv[arg + 1]);
			arg++;

			if (sigma_a < 0 || sigma_a > 10.0f)
			{
				PrintUsage();
			}
		}
		else if (_stricmp(argv[arg], "-sigma_t") == 0)
		{
			sigma_t = (float)atof(argv[arg + 1]);
			arg++;

			if (sigma_t < 0 || sigma_t > 10.0f)
			{
				PrintUsage();
			}
		}
		else
		{
			PrintUsage();
		}
	}
}

Config *InitConfig(int argc, const char *argv[])
{
	pConfig = new Config();
	pConfig->Parse(argc, argv);
	return pConfig;
}

Config *GetConfig() { return pConfig; }
