#pragma once

/*
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include <d3d11.h>
#include <DirectXPackedVector.h>
using namespace DirectX;

template <class T, int N> struct BufferVec
{
	T &operator[](UINT i) { return data[i]; }
	T data[N];
};

template <DXGI_FORMAT T> class CBuffType
{
};

template <> class CBuffType<DXGI_FORMAT_R32_FLOAT> : public BufferVec<float, 1>
{
};
template <> class CBuffType<DXGI_FORMAT_R16G16B16A16_FLOAT> : public BufferVec<DirectX::PackedVector::HALF, 4>
{
};
template <> class CBuffType<DXGI_FORMAT_R16_FLOAT> : public BufferVec<DirectX::PackedVector::HALF, 1>
{
};
template <> class CBuffType<DXGI_FORMAT_R8G8B8A8_UINT> : public BufferVec<BYTE, 4>
{
};
template <> class CBuffType<DXGI_FORMAT_R8G8_UINT> : public BufferVec<BYTE, 2>
{
};
template <> class CBuffType<DXGI_FORMAT_R8_UINT> : public BufferVec<BYTE, 1>
{
};

template <DXGI_FORMAT T> class CBuffSlice
{
public:
	CBuffSlice(UINT widthi, CBuffType<T> *data) : width(widthi), data(data) {}

	CBuffType<T> *operator[](UINT i) { return &data[i * width]; }

	UINT width;
	CBuffType<T> *data;
};

template <DXGI_FORMAT T> class CBuffer2D
{
public:
	CBuffer2D(UINT widthi, UINT heighti, UINT arraySizei)
	    : m_width(widthi), m_height(heighti), m_arraySize(arraySizei)
	{
	}

	~CBuffer2D() {}

protected:
	UINT m_width;
	UINT m_height;
	UINT m_arraySize;
};

template <DXGI_FORMAT T> class CBuffer2DRO : public CBuffer2D<T>
{
public:
	CBuffer2DRO(UINT width, UINT height, UINT arraySize) : CBuffer2D<T>(width, height, arraySize)
	{
		m_ptexture = NULL;
		UINT totalSize = width * height * arraySize;
		m_pbuff = new CBuffType<T>[totalSize];
		m_psr = new D3D11_SUBRESOURCE_DATA[arraySize];

		for (UINT i = 0; i < arraySize; i++)
		{
			m_psr[i].pSysMem = &m_pbuff[i * width * height];
			m_psr[i].SysMemPitch = width * sizeof(CBuffType<T>);
			m_psr[i].SysMemSlicePitch = 0;
		}

		m_pSRV = NULL;
	}

	CBuffSlice<T> operator[](UINT i) { return CBuffSlice<T>(m_width, &m_pbuff[i * m_width * m_height]); }

	void CreateGPUBuffers(ID3D11Device *pDevice)
	{
		if (!m_pSRV)
		{
			CreateTexture2D(pDevice, &m_ptexture);

			D3D11_SHADER_RESOURCE_VIEW_DESC desc;
			ZeroMemory(&desc, sizeof(desc));

			desc.Format = T;
			if (m_arraySize > 1)
			{
				desc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2DARRAY;
			}
			else
			{
				desc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
			}

			desc.Texture2DArray.ArraySize = m_arraySize;
			desc.Texture2DArray.FirstArraySlice = 0;
			desc.Texture2DArray.MipLevels = 1;
			desc.Texture2DArray.MostDetailedMip = 0;
			pDevice->CreateShaderResourceView(m_ptexture, &desc, &m_pSRV);
		}
	}

	~CBuffer2DRO()
	{
		SAFE_RELEASE(m_ptexture)
		SAFE_DELETE_ARRAY(m_pbuff);
		SAFE_DELETE_ARRAY(m_psr);
		SAFE_RELEASE(m_pSRV)
	}

	void reset()
	{
		UINT totalSize = sizeof(CBuffType<T>) * width * height * arraySize;
		memset(m_pbuff, 0, totalSize);
	}

	ID3D11ShaderResourceView *GetSRV() { return m_pSRV; }

protected:
	void CreateTexture2D(ID3D11Device *pDevice, ID3D11Texture2D **ppTexture)
	{
		D3D11_TEXTURE2D_DESC desc;
		ZeroMemory(&desc, sizeof(desc));

		desc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
		desc.Width = m_width;
		desc.Height = m_height;
		desc.Format = T;
		desc.MipLevels = 1;
		desc.ArraySize = m_arraySize;
		desc.SampleDesc.Count = 1;

		HRESULT hr = pDevice->CreateTexture2D(&desc, m_psr, ppTexture);
		if (FAILED(hr))
		{
			exit(1);
		}
	}

	ID3D11Texture2D *m_ptexture;
	D3D11_SUBRESOURCE_DATA *m_psr;
	ID3D11ShaderResourceView *m_pSRV;
	CBuffType<T> *m_pbuff;
};

template <DXGI_FORMAT T> class CBuffer2DRW : public CBuffer2D<T>
{
public:
	CBuffer2DRW(UINT width, UINT height, UINT arraySize) : CBuffer2D<T>(width, height, arraySize) 
	{ 
		m_pUAV = NULL; 
		m_pMappedBuffer = NULL;
	}

	~CBuffer2DRW()
	{
		SAFE_RELEASE(m_pUAV)
		SAFE_RELEASE(m_pBuffer)
		SAFE_RELEASE(m_pStagingBuffer)
	}

	void CreateGPUBuffers(ID3D11Device *pDevice)
	{
		if (!m_pUAV)
		{
			CreateStructuredBuffer(pDevice, &m_pBuffer, false);
			CreateStructuredBuffer(pDevice, &m_pStagingBuffer, true);

			D3D11_UNORDERED_ACCESS_VIEW_DESC desc;
			ZeroMemory(&desc, sizeof(desc));
			desc.ViewDimension = D3D11_UAV_DIMENSION_BUFFER;
			desc.Buffer.FirstElement = 0;
			desc.Buffer.NumElements = m_width * m_height * m_arraySize;
			desc.Format = DXGI_FORMAT_UNKNOWN;
			HRESULT hr = pDevice->CreateUnorderedAccessView(m_pBuffer, &desc, &m_pUAV);
			if (FAILED(hr))
			{
				exit(1);
			}
		}
	}

	CBuffSlice<T> operator[](UINT i) { return CBuffSlice<T>(m_width, &m_pMappedBuffer[i * m_width * m_height]); }

	ID3D11UnorderedAccessView *GetUAV() { return m_pUAV; }

	void CopyAndMap(ID3D11DeviceContext *pd3dImmediateContext)
	{
		pd3dImmediateContext->CopyResource(m_pStagingBuffer, m_pBuffer);
		HRESULT hr = pd3dImmediateContext->Map(m_pStagingBuffer, 0, D3D11_MAP_READ, 0, &m_MappedResource);
		if (FAILED(hr))
		{
			exit(1);
		}
		m_pMappedBuffer = (CBuffType<T> *)m_MappedResource.pData;
	}

	void Unmap(ID3D11DeviceContext *pd3dImmediateContext) { pd3dImmediateContext->Unmap(m_pStagingBuffer, 0); }

protected:
	void CreateStructuredBuffer(ID3D11Device *pDevice, ID3D11Buffer **ppBufOut, bool staging)
	{
		*ppBufOut = NULL;

		D3D11_BUFFER_DESC desc;
		ZeroMemory(&desc, sizeof(desc));
		desc.BindFlags = D3D11_BIND_UNORDERED_ACCESS | D3D11_BIND_SHADER_RESOURCE;
		desc.MiscFlags = D3D11_RESOURCE_MISC_BUFFER_STRUCTURED;
		desc.ByteWidth = sizeof(CBuffType<T>) * m_width * m_height * m_arraySize;
		desc.MiscFlags = D3D11_RESOURCE_MISC_BUFFER_STRUCTURED;
		desc.StructureByteStride = sizeof(CBuffType<T>);

		if (staging)
		{
			desc.BindFlags = 0;
			desc.MiscFlags = 0;
			desc.CPUAccessFlags = D3D11_CPU_ACCESS_READ;
			desc.Usage = D3D11_USAGE_STAGING;
		}

		HRESULT hr = pDevice->CreateBuffer(&desc, NULL, ppBufOut);
		if (FAILED(hr))
		{
			exit(1);
		}
	}

	ID3D11Buffer *m_pBuffer;
	ID3D11Buffer *m_pStagingBuffer;
	ID3D11UnorderedAccessView *m_pUAV;
	D3D11_MAPPED_SUBRESOURCE m_MappedResource;
	CBuffType<T> *m_pMappedBuffer;
};

template <class T> class CStructuredBuffer1DRW
{
public:
	CStructuredBuffer1DRW(UINT elements)
	{
		m_nElements = elements;
		m_pUAV = NULL;
		m_pSRV = NULL;
		m_pBuffer = NULL;
	}

	~CStructuredBuffer1DRW()
	{
		SAFE_RELEASE(m_pUAV)
		SAFE_RELEASE(m_pSRV)
		SAFE_RELEASE(m_pBuffer)
		m_nElements = 0;
	}

	void CreateGPUBuffers(ID3D11Device *pDevice)
	{
		if (!m_pBuffer)
		{
			D3D11_BUFFER_DESC desc;
			ZeroMemory(&desc, sizeof(desc));
			desc.Usage = D3D11_USAGE_DEFAULT;
			desc.BindFlags = D3D11_BIND_UNORDERED_ACCESS | D3D11_BIND_SHADER_RESOURCE;
			desc.MiscFlags = D3D11_RESOURCE_MISC_BUFFER_STRUCTURED;
			desc.ByteWidth = sizeof(T) * m_nElements;
			desc.StructureByteStride = sizeof(T);

			HRESULT hr = pDevice->CreateBuffer(&desc, NULL, &m_pBuffer);
			if (FAILED(hr))
			{
				exit(1);
			}
		}

		if (!m_pUAV)
		{
			D3D11_UNORDERED_ACCESS_VIEW_DESC desc;
			ZeroMemory(&desc, sizeof(desc));
			desc.Format = DXGI_FORMAT_UNKNOWN;
			desc.ViewDimension = D3D11_UAV_DIMENSION_BUFFER;
			desc.Buffer.FirstElement = 0;
			desc.Buffer.NumElements = m_nElements;
			HRESULT hr = pDevice->CreateUnorderedAccessView(m_pBuffer, &desc, &m_pUAV);
			if (FAILED(hr))
			{
				exit(1);
			}
		}

		if (!m_pSRV)
		{
			D3D11_SHADER_RESOURCE_VIEW_DESC desc;
			ZeroMemory(&desc, sizeof(desc));

			desc.Format = DXGI_FORMAT_UNKNOWN;
			desc.ViewDimension = D3D11_SRV_DIMENSION_BUFFER;
			desc.Buffer.ElementWidth = m_nElements;

			HRESULT hr = pDevice->CreateShaderResourceView(m_pBuffer, &desc, &m_pSRV);
			if (FAILED(hr))
			{
				exit(1);
			}
		}
	}

	ID3D11UnorderedAccessView *GetUAV() { return m_pUAV; }

	ID3D11ShaderResourceView *GetSRV() { return m_pSRV; }

protected:
	UINT m_nElements;
	ID3D11Buffer *m_pBuffer;
	ID3D11UnorderedAccessView *m_pUAV;
	ID3D11ShaderResourceView *m_pSRV;
};

template <class T> class CBufferConst
{
public:
	CBufferConst()
	{
		memset(&data, 0, sizeof(T));
		m_pBuffCB = NULL;
	}

	void CreateGPUBuffers(ID3D11Device *pDevice)
	{
		if (!m_pBuffCB)
		{
			// Create constant buffers
			D3D11_BUFFER_DESC Desc;
			ZeroMemory(&Desc, sizeof(Desc));

			Desc.Usage = D3D11_USAGE_DYNAMIC;
			Desc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
			Desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
			Desc.MiscFlags = 0;
			Desc.ByteWidth = sizeof(T);

			D3D11_SUBRESOURCE_DATA InitData;
			InitData.pSysMem = &data;
			HRESULT hr = pDevice->CreateBuffer(&Desc, &InitData, &m_pBuffCB);
			if (FAILED(hr))
			{
				exit(1);
			}
		}
	}

	~CBufferConst() { SAFE_RELEASE(m_pBuffCB); }

	ID3D11Buffer *GetBuffer() { return m_pBuffCB; }

	void UpdateBuffer(ID3D11DeviceContext *pd3dImmediateContext)
	{
		D3D11_MAPPED_SUBRESOURCE MappedResource;
		HRESULT hr = pd3dImmediateContext->Map(m_pBuffCB, 0, D3D11_MAP_WRITE_DISCARD, 0, &MappedResource);
		if (FAILED(hr))
		{
			exit(1);
		}
		memcpy(MappedResource.pData, &data, sizeof(T));
		pd3dImmediateContext->Unmap(m_pBuffCB, 0);
	}

	T * operator->() { return &data; }

	T data;

protected:
	ID3D11Buffer *m_pBuffCB;
};
