/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include <iostream>
#include <fstream>
#include "reconstruct.h"
#include "simpleBox.h"
#include "recon5D.h"
#include "recon5D_dx11.h"

// Algorithms
#include "reconstruct.h"
#include "simpleBox.h"

SampleInfo *g_pSamples = nullptr;
CReconstruct *g_pReconstruct = NULL;

void Cleanup()
{
	if (g_pSamples != NULL)
	{
		delete[] g_pSamples;
	}
	if (g_pReconstruct)
	{
		delete g_pReconstruct;
	}
}

void ComputeDepthRange(SampleInfo *pSamples, SampleInfoHeader *pSampleHeader, float &dmin, float &dmax)
{
	int xRes = pSampleHeader->xResolution;
	int yRes = pSampleHeader->yResolution;
	int spp = pSampleHeader->spp;
	dmin = FLT_MAX;
	dmax = -FLT_MAX;

	for (int i = 0; i < xRes; ++i)
	{
		for (int j = 0; j < yRes; ++j)
		{
			for (int k = 0; k < spp; ++k)
			{
				int idx = spp * (i + xRes * j) + k;
				SampleInfo &Si = pSamples[idx];

				if (Si.w < 1E-6)
					Si.w = 1E-6f;

				// Compute depth min/max
				if (Si.w > SCENE_MAX_DEPTH || Si.w < 0.0f)
				{
					continue; // Don't create layers beyond SCENE_MAX_DEPTH
				}

				dmin = std::min(dmin, Si.w);
				dmax = std::max(dmax, Si.w);
			}
		}
	}
}

/*
* This sample buffer reader uses a simplified version of
* A simplified Sample dump reader from the 
* "Temporal Light Field Reconstruction for Rendering Distribution Effects [Lehtinen et al, SIGGRAPH 2011]"
* source code release. Available at GitHub at:
* https://github.com/jiawen/temporal-lightfield-reconstruction/
* 
*/

void ReadSampleBuffer(const char *filename, SampleInfoHeader &SampleHeader)
{
	FILE *fp, *fph;
	errno_t err = fopen_s(&fp, filename, "rb");
	if (err != 0)
	{
		printf("File not openened \n");
		Config::PrintUsage();
	}

	err = fopen_s(&fph, (std::string(filename) + std::string(".header")).c_str(), "rt");
	if (err != 0)
	{
		printf("Header not openened \n");
		Config::PrintUsage();
	}

	bool separateHeader = (fph != NULL);
	if (!separateHeader)
	{
		fph = fp;
	}

	printf("Importing sample buffer... ");

	// Parse version and image size.
	float version;
	int width, height, numSamplesPerPixel;
	fscanf_s(fph, "Version %f\n", &version);
	fscanf_s(fph, "Width %d\n", &width);
	fscanf_s(fph, "Height %d\n", &height);
	fscanf_s(fph, "Samples per pixel %d\n", &numSamplesPerPixel);

	SampleHeader.spp = numSamplesPerPixel;
	SampleHeader.xResolution = width;
	SampleHeader.yResolution = height;

	// Parse the rest of the header.
	char motionModel[32];
	fscanf_s(fph, "Motion model: %s\n", motionModel, (unsigned int)_countof(motionModel));

	float cocCoeff[2];
	fscanf_s(fph, "CoC coefficients (coc radius = C0/w+C1): %f,%f\n", &cocCoeff[0], &cocCoeff[1]);
	SampleHeader.C0 = cocCoeff[0];
	SampleHeader.C1 = cocCoeff[1];

	bool binary = false;
	char encoding[32];
	if (fscanf_s(fph, "Encoding = %s\n", encoding, (unsigned int)_countof(encoding)) == 1)
	{
		binary = std::string(encoding) == std::string("binary");
	}

	char descriptor[64];
	fscanf_s(fph, "%s\n", descriptor, (unsigned int)_countof(descriptor)); // Unused

	g_pSamples = new SampleInfo[width * height * numSamplesPerPixel];

	// Parse samples.
	struct Entry
	{
		float x, y, z, w, u, v, t, r, g, b, a, mv_x, mv_y, mv_w, dwdx, dwdy;
	};

	// Read binary dump!
	const int num = width * height * numSamplesPerPixel;
	Entry *entries = new Entry[num];
	fread(entries, sizeof(Entry), num, fp);

	int sidx = 0;
	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
		{
			for (int k = 0; k < numSamplesPerPixel; ++k)
			{
				const Entry &e = entries[sidx++];
				int idx = numSamplesPerPixel * (x + width * y) + k;

				SampleInfo &Si = g_pSamples[idx];

				Si.imageX = e.x;
				Si.imageY = e.y;
				Si.w = e.w;
				Si.lensU = e.u;
				Si.lensV = -e.v;
				Si.time = e.t - 0.5f;
				Si.rgb[0] = e.r;
				Si.rgb[1] = e.g;
				Si.rgb[2] = e.b;
				Si.vx = e.mv_x;
				Si.vy = e.mv_y;
				Si.vz = e.mv_w;
			}
		}
	}

	delete[] entries;

	fclose(fp);
	if (separateHeader)
	{
		fclose(fph);
	}
	printf("done\n");
}

int main(int argc, const char *argv[])
{
	std::string FileName;
	std::ifstream SampleFile;

	if (argc < 2)
	{
		std::cerr << "No file specified" << std::endl;
		Config::PrintUsage();
		return 1;
	}

	// Read Configuration
	Config *pConfig = InitConfig(argc, argv);
	pConfig->Print();

	FileName += argv[argc - 1];

	SampleInfoHeader SampleHeader;
	ReadSampleBuffer(FileName.c_str(), SampleHeader);

	float dmin, dmax;
	ComputeDepthRange(g_pSamples, &SampleHeader, dmin, dmax);
	std::cout << "Depth min " << dmin << " , max " << dmax << std::endl;

	switch (pConfig->GetAlgo())
	{
	case RECON_GAUSS: // Simply Gauss weighted samples
		g_pReconstruct = new SimpleBox(g_pSamples, &SampleHeader);
		break;
	case RECON_5D: // Our algorithm
		g_pReconstruct = new Recon5D(g_pSamples, &SampleHeader, dmin, dmax);
		break;
	case RECON_5D_GPU: // Our algorithm GPU implementation
		g_pReconstruct = new Recon5DGPU(g_pSamples, &SampleHeader, dmin, dmax);
		break;
	default:
		std::cerr << "Invalid Reconstruction Mode";
		Config::PrintUsage();
		exit(0);
		break;
	}

	// Reconstruct
	std::cout << "Starting Reconstruction\n";
	float time = 0;
	Image *pImage = g_pReconstruct->ReconstructParallel(time);
	std::cout << "Finished Reconstruction in " << time << " ms \n";

	// Output Image
	size_t end = FileName.find_last_of('.');
	std::string OpFileName(FileName, 0, end);
	OpFileName += std::string("_") + pConfig->GetAlgoName() + ".png";

	std::cout << "Save file" << std::endl;

	const float exposure = 0.0f;
	const float gamma = 2.2f;
	pImage->SavePNG(OpFileName, exposure, gamma);

	// Cleanup
	Cleanup();

	return 0;
}
