/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include "partition.h"
#include "common.h"
#include <assert.h>
#include <algorithm>

///////////////////////////////// PARTITION CLASS ///////////////////////////////////////

Partition::Partition(float _rangeMin, float _rangeMax)
    : dmin(FLT_MAX)
    , dmax(-FLT_MAX)
    , smin(FLT_MAX)
    , smax(-FLT_MAX)
    , xsum(0)
    , x2sum(0)
    , ysum(0)
    , y2sum(0)
    , sampleWeight(0)
{
	rangeMin = _rangeMin;
	rangeMax = _rangeMax;
}

void Partition::Clear()
{
	dmin = FLT_MAX;
	dmax = -FLT_MAX;
	smin = FLT_MAX;
	smax = -FLT_MAX;
	xsum = 0;
	x2sum = 0;
	ysum = 0;
	y2sum = 0;
	sampleWeight = 0.0f;
}

void Partition::AddSample(const float depth, const float vx, const float vy, const float shear, const float weight)
{
	dmin = std::min(dmin, depth);
	dmax = std::max(dmax, depth);
	smin = std::min(smin, shear);
	smax = std::max(smax, shear);

	xsum += weight * vx;
	x2sum += weight * vx * vx;
	ysum += weight * vy;
	y2sum += weight * vy * vy;

	sampleWeight += weight;
}

inline float Partition::ComputeAngle(const float x, const float y) const
{
	float angle = 0;
	// Rotate such that the x-axis is aligned with the motion direction
	if (fabs(y) < EPS)
	{
		angle = x > 0 ? 0.0f : (float)M_PI;
	}
	else
	{
		angle = atan2(y, x);
	}

	return angle;
}

void Partition::ComputeFilter(const float clampVal, const float s_x, const float s_a, const float s_t)
{
	float FILTER_CLAMP = clampVal;

	// Constants
	const float sigma_x_pix = s_x;
	const float sigma_u_lens = s_a;
	const float sigma_t_shutter = s_t;

	const float sigma_x_by_u = sigma_x_pix / sigma_u_lens;
	const float sigma_x_by_t = sigma_x_pix / sigma_t_shutter;

	// Circle of confusion
	float s1 = smin;
	float s2 = smax;
	bool diffSignL = (s1 < 0) ^ (s2 < 0); // defocus
	float s1u = abs(s1);
	float s2u = abs(s2);
	float cmin = std::min(s1u, s2u); // cmin and cmax always positive
	float cmax = std::max(s1u, s2u);

	// Motion Blur Params
	float frac = 1.0f / sampleWeight;
	float xavg = frac * xsum;
	float yavg = frac * ysum;

	float fx = fabs(xavg);
	float fy = fabs(yavg);

	float std_avg_x = sqrt(fabs(frac * x2sum - xavg * xavg));
	float std_avg_y = sqrt(fabs(frac * y2sum - yavg * yavg));
	float radius = 1.0f * std::max(std_avg_x, std_avg_y);

	// Rotate such that the x-axis is aligned with the motion direction
	float angle = ComputeAngle(xavg, yavg);

	// Remap to rotated space
	float v_avg = sqrt(xavg * xavg + yavg * yavg);
	float vmin = std::max((float)EPS, v_avg - radius); // vmin and vmax always positive
	float vmax = v_avg + radius;

	float sigmax = sigma_x_pix;
	float sigmay = sigma_x_pix;
	float sigmat = sigma_t_shutter;
	float sigmau = sigma_u_lens;
	float sigmav = sigma_u_lens;
	float shearxt = 0.0f;
	float shearxu = 0.0f;
	float shearyv = 0.0f;

	// ======== (V,Y) FILTER ======================
	if (diffSignL && cmin >= sigma_x_by_u)
	{
		sigmay = sigma_x_pix;
		sigmav = sigma_u_lens;
	}
	else if (cmin < sigma_x_by_u)
	{
		shearyv = (s1 + s2) / 2.f;
		sigmay = sigma_x_pix;
		sigmav = 2.0f * sigma_x_pix / (cmax - cmin + EPS); //  2*Omega_u^max/H
	}
	else
	{
		// Derive Egan's filter
		float c_avg = 2.0f * s1 * s2 / (s1 + s2);
		float tsigmax = 2.0f * sigma_u_lens * (cmax * cmin) / (cmax - cmin + EPS);
		float tsigmau = sigma_u_lens;
		tsigmax = std::min(FILTER_CLAMP, tsigmax); // Clamp sigmax to avoid truncation artifacts

		// Flip the filter
		float as = tsigmax * tsigmax;
		float bs = tsigmau * tsigmau;
		float cs = c_avg * c_avg;
		float d = as + cs * bs;
		float f = tsigmax < EPS ? 1 + bs : 1 + bs * cs / as;
		sigmay = std::max(sigma_x_pix, fabs(tsigmau * c_avg / sqrt(f)));
		sigmav = sqrt(d / cs);
		shearyv = c_avg / f;
	}

	// ========  (X,U,T) FILTER ========================
	if (v_avg - radius < 0 && vmin >= sigma_x_by_t) // DoF Only
	{
		shearxt = 0;
		sigmat = sigma_t_shutter;
		sigmau = sigmav;
		sigmax = sigmay;
		shearxu = shearyv;
		angle = 0;
	}
	else if (vmin < sigma_x_by_t && cmin < sigma_x_by_u) // Cone intersects Omega_x = 1/(2PI*sigma_x_pix) plane
	{
		shearxt = (vmax + vmin) / 2.f;
		sigmat = 2.0f * sigma_x_pix / (vmax - vmin + EPS);
		sigmax = sigma_x_pix;
		sigmau = sigma_u_lens;
		shearxu = shearyv;
	}
	else if (vmin * sigma_t_shutter < cmin * sigma_u_lens) // Cone intersects Omega_u = 1/(2PI*sigma_u_lens) plane
	{
		sigmax = sigmay;
		sigmau = sigmav;
		shearxu = shearyv;

		shearxt = (vmax + vmin) / 2.f;
		shearxt = std::min(shearxt, FILTER_CLAMP);
		sigmat = sigma_u_lens * cmin / (vmax - shearxt + EPS);
		sigmat = std::max(sigmat, sigma_t_shutter);
	}
	else // Cone intersects Omega_t = sigma_t_shutter plane
	{
		v_avg = vmax > EPS ? 2.0f * vmin * vmax / (vmin + vmax) : 0;
		float tsigmat_x = 2.0f * sigma_t_shutter * (vmax * vmin) / (vmax - vmin + (float)EPS);
		tsigmat_x = std::min(FILTER_CLAMP, tsigmat_x); // Clamp sigmax to avoid truncation artifacts.

		float a2 = tsigmat_x * tsigmat_x;
		float b2 = sigma_t_shutter * sigma_t_shutter;
		float c2 = v_avg * v_avg;
		float d = a2 + c2 * b2;
		float f = 1 + b2 * c2 / a2;

		sigmax = std::max(sigma_x_pix, fabs(sigma_t_shutter * v_avg / sqrt(f)));
		sigmat = sqrt(d / c2);
		shearxt = v_avg / f;
		shearxu = (s1 + s2) / 2.f;
		shearxu = clamp(shearxu, -FILTER_CLAMP, FILTER_CLAMP);
		float c_avg = abs(shearxu);
		sigmau = sigma_t_shutter * vmin / (cmax - c_avg + EPS);
		sigmau = std::max(sigmau, sigma_u_lens);
	}

	//sigmax = sigma_x_pix;
	//sigmay = sigma_x_pix;
	//sigmat = sigma_t_shutter;
	//sigmau = sigma_u_lens;
	//sigmav = sigma_u_lens;

	fi.shearxt = shearxt;
	fi.sigmax = sigmax;
	fi.sigmay = sigmay;
	fi.sigmat = sigmat;
	fi.sigmau = sigmau;
	fi.sigmav = sigmav;
	fi.angle = angle;
	fi.shearxu = shearxu;
	fi.shearyv = shearyv;
}
