/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include "recon5D.h"
#include <iostream>

// Helper function to run a task
void TaskFunction(void *data, int taskIndex, int taskCount)
{
	Recon5D::TaskContext *tc = (Recon5D::TaskContext *)data;

	int bx = (taskIndex % (tc->tasksPerRow)) * tc->blockSize;
	int by = (taskIndex / (tc->tasksPerRow)) * tc->blockSize;

	tc->pRecon->ProcessBlock(bx, by);
}

///////////////////////////////// RECON5D CLASS ///////////////////////////////////////

Recon5D::Recon5D(const SampleInfo *pSamples, const SampleInfoHeader *pSampleHeader, const float dmin, const float dmax)
    : CReconstruct(pSamples, pSampleHeader)
{
	// Create depth ranges
	CreateDepthRanges(dmin, dmax);

	m_tilePadding = ((m_RecWin - m_BlockSize) / 2) & ~0x1;
};

Recon5D::~Recon5D() {}

Image *Recon5D::Reconstruct(float &time)
{
	unsigned long t_start = timeGetTime();

	// Parse Blocks
	// Ny,Nx
	int numblocks = 0;
	for (int i = 0; i < m_Height; i += m_BlockSize)
	{
		for (int j = 0; j < m_Width; j += m_BlockSize)
		{
			numblocks++;
			ProcessBlock(j, i);
		}
	}

	unsigned long t_end = timeGetTime();
	time = (float)(t_end - t_start);

	return &m_Image;
}

Image *Recon5D::ReconstructParallel(float &time)
{
	// Count nr of blocks
	int numtasks = 0;
	int tasksPerRow = 0;
	for (int j = 0; j < m_Width; j += m_BlockSize)
	{
		tasksPerRow++;
		for (int i = 0; i < m_Height; i += m_BlockSize)
		{
			numtasks++;
		}
	}

	TaskGroup tg;
	tg.AllocTaskInfo(numtasks);

	// Reuse one task context for all tasks
	TaskContext *td = new TaskContext();
	td->blockSize = m_BlockSize;
	td->tasksPerRow = tasksPerRow;
	td->pRecon = this;

	// Create tasks
	for (int i = 0; i < numtasks; ++i)
	{
		TaskInfo *ti = tg.GetTaskInfo(i);
		ti->data = td;
		ti->func = TaskFunction;
		ti->taskIndex = i;
		ti->taskCount = numtasks;
	}

	// Launch Tasks
	unsigned long t_start = timeGetTime();

	tg.Launch(numtasks);
	tg.Sync();

	unsigned long t_end = timeGetTime();
	time = (float)(t_end - t_start);

	delete td;

	return &m_Image;
}

float Recon5D::findNextCoC(float z)
{
	static const float eps = 0.0001f;
	static const float CoCSplits[] = {-16.50f, -11.00f, -8.00f, -6.50f, -5.50f, -4.50f, -4.00f, -3.50f, -3.00f, -2.50f,
	                                  -2.00f,  -1.50f,  -1.00f, -0.5f,  0.0f,   0.5f,   1.00f,  1.50f,  2.00f,  2.50f,
	                                  3.00f,   3.50f,   4.00f,  4.50f,  5.50f,  6.50f,  8.00f,  11.00f, 16.50f};
	static const int NUM_COC_SPLITS = sizeof(CoCSplits) / sizeof(float);

	float c = m_Camera.ShearUV(z)
	          + eps; // Adding an epsilon to avoid getting stuck on CoC borders due to floating point precision issues
	for (int i = 0; i < NUM_COC_SPLITS; ++i)
	{
		if (c < CoCSplits[i])
		{
			return m_Camera.ShearUVInv(CoCSplits[i]);
		}
	}
	return FLT_MAX;
}

// Partition scene in depth layers
void Recon5D::CreateDepthRanges(float dmin, float dmax)
{
	Config *pConfig = GetConfig();
	float dPrev = dmin;

	if (!m_Camera.hasDof())
	{
		// Camera has no DOF, just do uniform splitting based on depths
		const float layerScale = (dmax - dmin) / (float)pConfig->DepthLayers;
		for (int p = 0; p < pConfig->DepthLayers; p++)
		{
			m_depthRanges.push_back(float2(dmin + p * layerScale, dmin + (p + 1) * layerScale));
		}
	}
	else
	{
		// Camera has DOF, so we use the CoC splitting lookup table. Also, use a
		// maximum allowed depth step, and use that step if the CoC is "too large"
		float maxStep = (dmax - dmin) / (float)pConfig->DepthLayers;
		while (dPrev < dmax)
		{
			float dNext = findNextCoC(dPrev);

			if (dNext - dPrev > maxStep)
			{
				dNext = dPrev + std::min(maxStep, (dNext + dPrev) / 2.0f);
			}

			m_depthRanges.push_back(float2(dPrev, dNext));
			dPrev = dNext;
		}
	}

	// Insert a layer from [dmax, FLT_MAX] to catch depths of cleared pixels
	m_depthRanges.push_back(float2(dPrev, FLT_MAX));
}

inline void Recon5D::ProcessBlock(int bx, int by)
{
	int Xmin = bx - m_tilePadding;
	int Ymin = by - m_tilePadding;
	int Xmax = bx + m_BlockSize + m_tilePadding;
	int Ymax = by + m_BlockSize + m_tilePadding;

	TileData *td = new TileData(m_RecWin, m_depthRanges);

	ComputeLayers(td, Xmin, Ymin, Xmax, Ymax);

	const float filterClampVal = m_tilePadding / 2.0f;
	for (auto &p : td->partList)
	{
		p.ComputeFilter(filterClampVal, m_sigma_x, m_sigma_a, m_sigma_t);
	}

	for (auto &part : td->partList)
	{
		td->filter->Reset();
		td->filter1->Reset();
		Reproject(td, part, Xmin, Ymin, Xmax, Ymax);
		FilterPartition(td, part);
		ComputeColorPerPixel(td, bx, by);
	}

	delete td;
}

// This function assumes that the layers are sorted front to back
void Recon5D::ComputeColorPerPixel(TileData *td, int Xmin, int Ymin)
{
	int offset = m_tilePadding;
	int BlockX = std::min(Xmin + m_BlockSize, m_Width) - Xmin;
	int BlockY = std::min(Ymin + m_BlockSize, m_Height) - Ymin;

	float4 *CBuffer = m_Image.GetBuffer();

	for (int i = 0; i < BlockY; i++)
	{
		for (int j = 0; j < BlockX; j++)
		{
			int idx = ((Ymin + i) * m_Width + (Xmin + j));

			PixelInfo p = td->filter->FetchData(j + offset, i + offset);
			float &vis = td->filterVis->FetchData(j + offset, i + offset); // This is updated

			float OneOverWeight = 1 / (p.w + EPS);
			float alpha = p.col.get(3) * OneOverWeight; // opacity

			CBuffer[idx] += p.col*(vis * OneOverWeight);
			vis *= 1.0f - alpha;
		}
	}
}

PixelInfo Recon5D::FilteredRead(float x, float y, CFilter<PixelInfo> *flt) 
{
	if (GetConfig()->LookupFilter == FILTER_NEAREST)
	{
		return flt->FetchData( (int)(x + 0.5f), (int)(y + 0.5f) );
	}
	else
	{	
		int x0 = (int)x;
		int y0 = (int)y;
		int x1 = std::min(m_RecWin - 1, x0 + 1);
		int y1 = std::min(m_RecWin - 1, y0 + 1);

		PixelInfo &p00 = flt->FetchData(x0, y0);
		PixelInfo &p10 = flt->FetchData(x1, y0);
		PixelInfo &p01 = flt->FetchData(x0, y1);
		PixelInfo &p11 = flt->FetchData(x1, y1);

		float sx = x - (float)(x0);
		float sy = y - (float)(y0);

		PixelInfo px0 = p00 * (1.f - sx) + p10 * sx;
		PixelInfo px1 = p01 * (1.f - sx) + p11 * sx;

		return px0 * (1.f - sy) + px1 * sy;
	}
}

// Separable, rotated anisotropic filter in XY
void Recon5D::FilterPartition(TileData *td, const Partition &part)
{
	FilterInfo fi = part.GetFilterInfo();

	const float f_RecWin = (float)m_RecWin - 1.0f;
	float scale = -0.5f / (fi.sigmax * fi.sigmax);
	int radius = std::min(int(fi.sigmax * 3.0f), m_tilePadding);
	float c_a = cos(fi.angle);
	float s_a = sin(fi.angle);

	// Filter X
	for (int k = -radius; k <= radius; k++)
	{
		float w = exp(k * k * scale);

		for (int i = 0; i < m_RecWin; i++)
		{
			for (int j = 0; j < m_RecWin; j++)
			{
				// Apply rotation: x' = R x
				float x = j + k * c_a;
				float y = i - k * s_a;

				// Check if rotated filter tap is outside the tile
				if (x < 0.0f || x >= f_RecWin || y < 0.0f || y >= f_RecWin)
					continue;

				PixelInfo pi = FilteredRead(x, y, td->filter);
				PixelInfo &pic = td->filter1->FetchData(j, i);
				pic += pi * w;
			}
		}
	}

	// Filter Y
	scale = -0.5f / (fi.sigmay * fi.sigmay); // filter width in axis orthogonal to motion vector. This is narrow unless there is DOF
	radius = std::min(int(fi.sigmay * 3.0f), m_tilePadding);
	td->filter->Reset();

	for (int k = -radius; k <= radius; k++)
	{
		float w = exp(k * k * scale);
		for (int i = 0; i < m_RecWin; i++)
		{
			for (int j = 0; j < m_RecWin; j++)
			{
				float x = j + k * s_a;
				float y = i + k * c_a;

				// Check if rotated filter tap is outside the tile
				if (x < 0.0f || x >= f_RecWin || y < 0.0f || y >= f_RecWin)
					continue;

				PixelInfo pi = FilteredRead(x, y, td->filter1);
				PixelInfo &pic = td->filter->FetchData(j, i);
				pic += pi * w;
			}
		}
	}
}

void Recon5D::ComputeLayers(TileData *td, int Xmin, int Ymin, int Xmax, int Ymax)
{
	int Xminc = std::max(0, Xmin);
	int Yminc = std::max(0, Ymin);
	int Xmaxc = std::min(m_Width, Xmax);
	int Ymaxc = std::min(m_Height, Ymax);

	float tileCenterX = (Xmin + Xmax) / 2.0f;
	float tileCenterY = (Ymin + Ymax) / 2.0f;

	float OneOverSigma2 = 1.f / (m_BlockSize * m_BlockSize);

	// Build histogram of samples, i.e., fill layers
	for (int i = Yminc; i < Ymaxc; i++)
	{
		for (int j = Xminc; j < Xmaxc; j++)
		{
			for (int k = 0; k < m_spp; ++k)
			{
				int idx = m_spp * (j + m_Width * i) + k;

				const SampleInfo &si = m_pSamples[idx];
				float shear = m_Camera.ShearUVClamp(si.w);

				// Sample weight
				float dx = si.imageX - tileCenterX;
				float dy = si.imageY - tileCenterY;
				float r2 = dx * dx + dy * dy;

				float weight = exp(-r2 * OneOverSigma2);

				// Find depth partition
				bool foundPart = false;
				for (auto &p : td->partList)
				{
					if (p.InRange(si.w))
					{
						float2 mv = m_Camera.PixelMotion(si);
						p.AddSample(si.w, mv.x, mv.y, shear, weight);
						foundPart = true;
						break;
					}
				}

				if (!foundPart)
				{
					printf("WARNING: Sample didn't get assigned to a partition (%f)\n", si.w);
				}
			}
		}
	}

	// Compact partition list
	td->partList.remove_if([](Partition &p)
	                       {
		                       return p.IsEmpty();
		                   });
}

void Recon5D::FilteredSplat(TileData *td, float x, float y, const float4 col, float w)
{
	if (GetConfig()->LookupFilter == FILTER_NEAREST)
	{
		int xi = (int)x;
		int yi = (int)y;

		PixelInfo &pi = td->filter->FetchData(xi, yi);
		pi.col += col * w;
		pi.w   += w;
	}
	else
	{
		x = std::max(0.0f, std::min((float)m_RecWin, x - 0.5f));
		y = std::max(0.0f, std::min((float)m_RecWin, y - 0.5f));

		int x0 = (int)x;
		int y0 = (int)y;
		int x1 = std::min(m_RecWin - 1, x0 + 1);
		int y1 = std::min(m_RecWin - 1, y0 + 1);

		PixelInfo &p_00 = td->filter->FetchData(x0, y0);
		PixelInfo &p_10 = td->filter->FetchData(x1, y0);
		PixelInfo &p_01 = td->filter->FetchData(x0, y1);
		PixelInfo &p_11 = td->filter->FetchData(x1, y1);

		float w_x = x - (float)(x0);
		float w_y = y - (float)(y0);

		float w_00 = w * (1 - w_x) * (1 - w_y);
		float w_10 = w * (w_x) * (1 - w_y);
		float w_01 = w * (1 - w_x) * (w_y);
		float w_11 = w * (w_x) * (w_y);

		p_00.col += col * w_00;
		p_10.col += col * w_10;
		p_01.col += col * w_01;
		p_11.col += col * w_11;

		p_00.w += w_00;
		p_10.w += w_10;
		p_01.w += w_01;
		p_11.w += w_11;
	}
}

void Recon5D::FilteredSplatWeight(TileData *td, float x, float y, float w)
{
	if (GetConfig()->LookupFilter == FILTER_NEAREST)
	{
		int xi = (int)(x + 0.5f);
		int yi = (int)(y + 0.5f);

		PixelInfo &pi = td->filter->FetchData(xi, yi);
		pi.w   += w;
	}
	else
	{
		x = std::max(0.0f, std::min((float)m_RecWin, x));
		y = std::max(0.0f, std::min((float)m_RecWin, y));

		int x0 = (int)x;
		int y0 = (int)y;
		int x1 = std::min(m_RecWin - 1, x0 + 1);
		int y1 = std::min(m_RecWin - 1, y0 + 1);

		PixelInfo &p_00 = td->filter->FetchData(x0, y0);
		PixelInfo &p_10 = td->filter->FetchData(x1, y0);
		PixelInfo &p_01 = td->filter->FetchData(x0, y1);
		PixelInfo &p_11 = td->filter->FetchData(x1, y1);

		float w_x = x - (float)(x0);
		float w_y = y - (float)(y0);

		p_00.w += w * (1 - w_x) * (1 - w_y);
		p_10.w += w * (w_x) * (1 - w_y);
		p_01.w += w * (1 - w_x) * (w_y);
		p_11.w += w * (w_x) * (w_y);
	}
}

void Recon5D::Reproject(TileData *td, const Partition &part, int Xmin, int Ymin, int Xmax, int Ymax)
{
	int Xminc = std::max(0, Xmin);
	int Yminc = std::max(0, Ymin);
	int Xmaxc = std::min(m_Width, Xmax);
	int Ymaxc = std::min(m_Height, Ymax);

	FilterInfo fi = part.GetFilterInfo();
	float s_a = sin(fi.angle);
	float c_a = cos(fi.angle);

	// Rotate motion vector:
	float vx_r = c_a * fi.shearxt;
	float vy_r = s_a * fi.shearxt;

	// Rotate defocus parameters:
	float a00 = c_a * c_a * fi.shearxu + s_a * s_a * fi.shearyv;
	float a01 = s_a * c_a * (fi.shearxu - fi.shearyv);
	float a10 = a01;
	float a11 = c_a * c_a * fi.shearyv + s_a * s_a * fi.shearxu;

	const float uweight = -0.5f / (fi.sigmau * fi.sigmau);
	const float vweight = -0.5f / (fi.sigmav * fi.sigmav);
	const float tweight = -0.5f / (fi.sigmat * fi.sigmat);
	const float f_RecWin = (float)m_RecWin - 1.0f;

	// Traverse every pixel in search window
	for (int i = Yminc; i < Ymaxc; i++)
	{
		for (int j = Xminc; j < Xmaxc; j++)
		{
			for (int k = 0; k < m_spp; k++)
			{
				// Get sample
				int idx = m_spp * (j + m_Width * i) + k;
				const SampleInfo &si = m_pSamples[idx];
				if (part.IsFullyBehindSample(si.w))
				{
					continue;
				}

				// Reproject to the center of the lens and t=0 (center of shutter) and accumulate
				float x0 = si.imageX - Xmin - a00 * si.lensU - a01 * si.lensV - vx_r * si.time;
				float y0 = si.imageY - Ymin + a10 * si.lensU + a11 * si.lensV + vy_r * si.time;

				if (x0 <= 0.0f || x0 >= f_RecWin || y0 <= 0.0f || y0 >= f_RecWin)
					continue;

				float w
				    = exp(uweight * si.lensU * si.lensU + vweight * si.lensV * si.lensV + tweight * si.time * si.time);

				if (part.InTrimmedRange(si.w))
				{
					FilteredSplat(td, x0, y0, float4(si.rgb[0], si.rgb[1], si.rgb[2], 1.0f), w);
				}
				else if (part.IsInFrontOfSample(si.w))
				{
					FilteredSplatWeight(td, x0, y0, w);
				}
			}
		}
	}
}