#pragma once

/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include "reconstruct.h"
#include "tiledata.h"

void TaskFunction(void *data, int taskIndex, int taskCount);

class Recon5D : public CReconstruct
{
public:
	Recon5D(const SampleInfo *pSamples, const SampleInfoHeader *pSampleHeader, const float dmin, const float dmax);

	virtual ~Recon5D();

	virtual Image *Reconstruct(float &time);
	virtual Image *ReconstructParallel(float &time);

	// Helper function to launch asyncronous tasks
	friend void TaskFunction(void *data, int taskIndex, int taskCount);

private:
	struct TaskContext
	{
		Recon5D *pRecon;
		int blockSize;
		int tasksPerRow;
	};

	int m_tilePadding;
	std::vector<float2> m_depthRanges; // Holds limits of static depth partitions (scene dependent)

	//
	// Helper function to insert uniformly distributed depth layers
	//
	void CreateUniformDepthRanges(const int nLayers, const float dmin, const float dmax);

	float findNextCoC(float z);

	//
	// Divide the scene in depth layers. Add more layers close to the plane in focus
	//
	void CreateDepthRanges(const float dmin, const float dmax);

	//
	// Main method to filter one tile:
	// 1. Divide the scene in depth layers
	// 2. Build histogram over samples and compute filter parameters
	// 3. For each layer (front to back)
	//		- Integrate over uvt (multiplication and projection of samples)
	//      - Apply screen space convolution
	//      - Add to frame buffer using alpha blending
	void ProcessBlock(int bx, int by);

	//
	// Perform a filtered pixel splatting operation , using either nearest 
	// or bilinear filtering
	//
	void FilteredSplat(TileData *td, float x, float y, const float4 col, float w);

	//
	// Perform a filtered pixel splatting operation , using either nearest 
	// or bilinear filtering. Weight only
	//
	void FilteredSplatWeight(TileData *td, float x, float y, float w);

	//
	// Build histogram over samples. Add samples to depth layers, and compute
	// filter parameters
	//
	void ComputeLayers(TileData *td, int Xmin, int Ymin, int Xmax, int Ymax);

	//
	// Perform integration over (u,v,t). Implemented as a per-sample
	// multiplication followed by a oblique projection down to the (x,y) plane
	//
	void Reproject(TileData *td, const Partition &part, int Xmin, int Ymin, int Xmax, int Ymax);

	//
	// Perform a filtered buffer read, using either nearest or bilinear filtering
	//
	PixelInfo FilteredRead(float x, float y, CFilter<PixelInfo> *flt);

	//
	// Apply a screen space roated Gaussian filter The filter is separable in x and y
	//
	void FilterPartition(TileData *td, const Partition &part);

	//
	// Composite a layer using a modified alpha blending
	//
	void ComputeColorPerPixel(TileData *td, int Xmin, int Ymin);

};
