/* 
* Copyright (c) 2013-2014, Intel Corporation
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* - Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Neither the name of Intel Corporation nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include <d3d11.h>
#include <DirectXPackedVector.h>
#include <d3dcompiler.h>
#include <windows.h>

#undef min
#undef max

#include "dx11_utils/measurement.h"
#include "../shaders/tileconfig.h"

#include "recon5D_dx11.h"

///////////////////////////////// RECON5D_GPU CLASS ///////////////////////////////////////

Recon5DGPU::Recon5DGPU(const SampleInfo *pSamples, const SampleInfoHeader *pSampleHeader, const float dmin,
                       const float dmax)
    : CReconstruct(pSamples, pSampleHeader)
{
	printf("Creating device...");
	m_pDevice = NULL;
	m_pContext = NULL;
	CreateComputeDevice();
	printf("done\n");

	printf("Loading compute shader...");
	CreateComputeShader(L"reconstruction.cso", &m_pReconstructCS);
	CreateComputeShader(L"layers.cso", &m_pLayerStatsCS);
	printf("done\n");

	m_pImageSettings = new CBufferConst<SImageSettings>();

	SImageSettings &imgSettings = m_pImageSettings->data;
	imgSettings.Nx = pSampleHeader->xResolution;
	imgSettings.Ny = pSampleHeader->yResolution;
	imgSettings.Tx = (imgSettings.Nx + IMAGE_INNER_TILE - 1) / IMAGE_INNER_TILE;
	imgSettings.Ty = (imgSettings.Ny + IMAGE_INNER_TILE - 1) / IMAGE_INNER_TILE;
	imgSettings.spp = pSampleHeader->spp;
	imgSettings.DepthMin = dmin;
	imgSettings.DepthMax = dmax;
	imgSettings.MaxBlur = GetConfig()->MaxMotion;
	imgSettings.sigma_x = GetConfig()->sigma_x;
	imgSettings.sigma_a = GetConfig()->sigma_a;
	imgSettings.sigma_t = GetConfig()->sigma_t;
	imgSettings.Camera_C0 = m_Camera.GetC0();
	imgSettings.Camera_C1 = m_Camera.GetC1();

	// Setup GPU side sample buffers
	printf("Creating GPU sample data buffers...");
	m_pBufXYUV = new CBuffer2DRO<DXGI_FORMAT_R8G8B8A8_UINT>(imgSettings.Nx, imgSettings.Ny, imgSettings.spp);
	m_pBufT = new CBuffer2DRO<DXGI_FORMAT_R8_UINT>(imgSettings.Nx, imgSettings.Ny, imgSettings.spp);
	m_pBufVel = new CBuffer2DRO<DXGI_FORMAT_R8G8_UINT>(imgSettings.Nx, imgSettings.Ny, imgSettings.spp);
	m_pBuffC = new CBuffer2DRO<DXGI_FORMAT_R16G16B16A16_FLOAT>(imgSettings.Nx, imgSettings.Ny, imgSettings.spp);
	m_pBuffZ = new CBuffer2DRO<DXGI_FORMAT_R32_FLOAT>(imgSettings.Nx, imgSettings.Ny, imgSettings.spp);
	m_pBufComposite = new CBuffer2DRW<DXGI_FORMAT_R16G16B16A16_FLOAT>(imgSettings.Nx, imgSettings.Ny, 1);
	m_pBufLayers = new CStructuredBuffer1DRW<SLayerInfo>(imgSettings.Tx * imgSettings.Ty * MAX_LAYERS);

	// Write sample data into GPU buffers and convert to appropriate formats
	for (UINT y = 0; y < imgSettings.Ny; y++)
	{
		for (UINT x = 0; x < imgSettings.Nx; x++)
		{
			for (UINT s = 0; s < imgSettings.spp; s++)
			{
				UINT idx = m_spp * (x + imgSettings.Nx * y) + s;
				const SampleInfo &si = m_pSamples[idx];

				CBuffType<DXGI_FORMAT_R8G8B8A8_UINT> &xyuv = (*m_pBufXYUV)[s][y][x];
				xyuv[0] = (BYTE)((si.imageX - x) * 255.f);
				xyuv[1] = (BYTE)((si.imageY - y) * 255.f);
				xyuv[2] = (BYTE)((si.lensU + 1) * 127.f);
				xyuv[3] = (BYTE)((si.lensV + 1) * 127.f);

				CBuffType<DXGI_FORMAT_R8_UINT> &t = (*m_pBufT)[s][y][x];
				t[0] = (BYTE)((si.time + 0.5f) * 255.f);

				CBuffType<DXGI_FORMAT_R8G8_UINT> &vel = (*m_pBufVel)[s][y][x];
				float2 fvel = m_Camera.PixelMotion(si);
				FLOAT vx = 127.5f + fvel.x * 127.5f / imgSettings.MaxBlur;
				FLOAT vy = 127.5f + fvel.y * 127.5f / imgSettings.MaxBlur;

				vel[0] = (BYTE)std::max(0.0f, std::min(255.0f, vx));
				vel[1] = (BYTE)std::max(0.0f, std::min(255.0f, vy));

				CBuffType<DXGI_FORMAT_R16G16B16A16_FLOAT> &rgba = (*m_pBuffC)[s][y][x];
				rgba[0] = PackedVector::XMConvertFloatToHalf(si.rgb[0]);
				rgba[1] = PackedVector::XMConvertFloatToHalf(si.rgb[1]);
				rgba[2] = PackedVector::XMConvertFloatToHalf(si.rgb[2]);
				rgba[3] = PackedVector::XMConvertFloatToHalf(1);
				CBuffType<DXGI_FORMAT_R32_FLOAT> &z = (*m_pBuffZ)[s][y][x];
				z[0] = si.w;
			}
		}
	}

	// Submit buffers
	m_pBufXYUV->CreateGPUBuffers(m_pDevice);
	m_pBufT->CreateGPUBuffers(m_pDevice);
	m_pBufVel->CreateGPUBuffers(m_pDevice);
	m_pBuffC->CreateGPUBuffers(m_pDevice);
	m_pBuffZ->CreateGPUBuffers(m_pDevice);
	m_pBufComposite->CreateGPUBuffers(m_pDevice);
	m_pImageSettings->CreateGPUBuffers(m_pDevice);
	m_pBufLayers->CreateGPUBuffers(m_pDevice);
	printf("done\n");
}

Recon5DGPU::~Recon5DGPU()
{
	SAFE_DELETE(m_pBufComposite);
	SAFE_DELETE(m_pBuffC);
	SAFE_DELETE(m_pBuffZ);
	SAFE_DELETE(m_pBufXYUV);
	SAFE_DELETE(m_pBufT);
	SAFE_DELETE(m_pBufVel);
	SAFE_DELETE(m_pImageSettings);
	SAFE_DELETE(m_pBufLayers);

	SAFE_RELEASE(m_pLayerStatsCS);
	SAFE_RELEASE(m_pReconstructCS);
	SAFE_RELEASE(m_pContext);
	SAFE_RELEASE(m_pDevice);
}

Image *Recon5DGPU::Reconstruct(float &time) { return ReconstructParallel(time); }

Image *Recon5DGPU::ReconstructParallel(float &time)
{
	ID3D11ShaderResourceView *aRViews[] = {m_pBuffC->GetSRV(), m_pBuffZ->GetSRV(),  m_pBufXYUV->GetSRV(),
	                                       m_pBufT->GetSRV(),  m_pBufVel->GetSRV(), m_pBufLayers->GetSRV()};
	ID3D11ShaderResourceView *aRViewsLayers[]
	    = {m_pBuffC->GetSRV(), m_pBuffZ->GetSRV(), m_pBufXYUV->GetSRV(), m_pBufT->GetSRV(), m_pBufVel->GetSRV()};

	ID3D11UnorderedAccessView *aUAViews[] = {m_pBufComposite->GetUAV()};
	ID3D11UnorderedAccessView *aUAViewsLayers[] = {m_pBufLayers->GetUAV()};

	SImageSettings &imgSettings = m_pImageSettings->data;
	UINT X = (imgSettings.Nx + IMAGE_INNER_TILE - 1) / IMAGE_INNER_TILE;
	UINT Y = (imgSettings.Ny + IMAGE_INNER_TILE - 1) / IMAGE_INNER_TILE;

	CMeasurement timing(m_pDevice);

	timing.BeginMeasurement(m_pContext);
	RunComputeShader(m_pLayerStatsCS, sizeof(aRViewsLayers) / sizeof(ID3D11ShaderResourceView *), aRViewsLayers,
	                 sizeof(aUAViewsLayers) / sizeof(ID3D11UnorderedAccessView *), aUAViewsLayers, X, Y);
	RunComputeShader(m_pReconstructCS, sizeof(aRViews) / sizeof(ID3D11ShaderResourceView *), aRViews,
	                 sizeof(aUAViews) / sizeof(ID3D11UnorderedAccessView *), aUAViews, X, Y);
	time = timing.EndMeasurement(m_pContext);

	CopyOutput();
	return &m_Image;
}

///////////////////////////////// Helper functions ///////////////////////////////////////

void Recon5DGPU::CopyOutput()
{
	m_pBufComposite->CopyAndMap(m_pContext);

	SImageSettings &imgSettings = m_pImageSettings->data;
	float4 *CBuffer = m_Image.GetBuffer();

	for (UINT y = 0; y < imgSettings.Ny; ++y)
	{
		for (UINT x = 0; x < imgSettings.Nx; ++x)
		{
			// Read pixel & convert to float
			float rgba[4];
			for (UINT c = 0; c < 4; ++c)
			{
				rgba[c] = PackedVector::XMConvertHalfToFloat((*m_pBufComposite)[0][y][x][c]);
			}

			// Normalize & write to output struct
			//CBuffer[x + y * imgSettings.Nx].x = rgba[0] / (float)COLOR_SCALE;
			//CBuffer[x + y * imgSettings.Nx].y = rgba[1] / (float)COLOR_SCALE;
			//CBuffer[x + y * imgSettings.Nx].z = rgba[2] / (float)COLOR_SCALE;
			//CBuffer[x + y * imgSettings.Nx].w = 1.0f;

			CBuffer[x + y * imgSettings.Nx] = float4(rgba[0] / (float)COLOR_SCALE, rgba[1] / (float)COLOR_SCALE, rgba[2] / (float)COLOR_SCALE, 1.0f);

		}
	}

	m_pBufComposite->Unmap(m_pContext);
}

void Recon5DGPU::RunComputeShader(ID3D11ComputeShader *pComputeShader, UINT nNumViews,
                                  ID3D11ShaderResourceView **pShaderResourceViews, UINT nNumUAVs,
                                  ID3D11UnorderedAccessView **pUnorderedAccessView, UINT X, UINT Y)
{
	m_pContext->CSSetShader(pComputeShader, NULL, 0);
	m_pContext->CSSetShaderResources(0, nNumViews, pShaderResourceViews);
	m_pContext->CSSetUnorderedAccessViews(0, nNumUAVs, pUnorderedAccessView, NULL);

	ID3D11Buffer *pBuffCB = m_pImageSettings->GetBuffer();
	m_pContext->CSSetConstantBuffers(0, 1, &pBuffCB);
	m_pContext->Dispatch(X, Y, 1);

	// Cleanup
	ID3D11UnorderedAccessView *ppUAViewNULL[] = {NULL};
	ID3D11ShaderResourceView *ppSRVNULL[] = {NULL, NULL, NULL, NULL, NULL, NULL};
	ID3D11Buffer *ppCBNULL[] = {NULL};

	m_pContext->CSSetShader(NULL, NULL, 0);
	m_pContext->CSSetUnorderedAccessViews(0, sizeof(ppUAViewNULL) / sizeof(ID3D11UnorderedAccessView *), ppUAViewNULL,
	                                      NULL);
	m_pContext->CSSetShaderResources(0, sizeof(ppSRVNULL) / sizeof(ID3D11ShaderResourceView *), ppSRVNULL);
	m_pContext->CSSetConstantBuffers(0, sizeof(ppCBNULL) / sizeof(ID3D11Buffer *), ppCBNULL);
}

void Recon5DGPU::CreateComputeDevice()
{
	// Load d3d11.dll
	HMODULE hModD3D11 = LoadLibrary(L"d3d11.dll");
	if (hModD3D11 == NULL)
	{
		printf("Failed to load d3d11.dll\n");
		exit(1);
	}

	// Get entry point to D3D11CreateDevice
	typedef HRESULT(WINAPI * LPD3D11CREATEDEVICE)(IDXGIAdapter *, D3D_DRIVER_TYPE, HMODULE, UINT32,
	                                              CONST D3D_FEATURE_LEVEL *, UINT, UINT32, ID3D11Device **,
	                                              D3D_FEATURE_LEVEL *, ID3D11DeviceContext **);
	LPD3D11CREATEDEVICE DynamicD3D11CreateDevice = (LPD3D11CREATEDEVICE)GetProcAddress(hModD3D11, "D3D11CreateDevice");
	if (DynamicD3D11CreateDevice == NULL)
	{
		printf("Failed to load d3d11.dll\n");
		exit(1);
	}

	// Create d3d11 device
	D3D_FEATURE_LEVEL featureLevelOut;
	const D3D_FEATURE_LEVEL featureLevel[] = {D3D_FEATURE_LEVEL_11_0};

	HRESULT hr = DynamicD3D11CreateDevice(NULL,                     // Use default graphics card
	                                      D3D_DRIVER_TYPE_HARDWARE, // Create a hardware accelerated device
	                                      NULL,                     // Do not use external software rasterizer module
#if defined(DEBUG) || defined(_DEBUG)
	                                      D3D11_CREATE_DEVICE_SINGLETHREADED | D3D11_CREATE_DEVICE_DEBUG,
#else
	                                      D3D11_CREATE_DEVICE_SINGLETHREADED,
#endif
	                                      featureLevel, 
										  sizeof(featureLevel) / sizeof(D3D_FEATURE_LEVEL),
	                                      D3D11_SDK_VERSION,		// SDK version
	                                      &m_pDevice,				// Device out
	                                      &featureLevelOut,			// Actual feature level created
	                                      &m_pContext);				// Context out

	// Check that we got a valid d3d11 device
	if (FAILED(hr) || featureLevelOut != D3D_FEATURE_LEVEL_11_0)
	{
		printf("Failed to create DirectX 11 device\n");
		exit(1);
	}
}

void Recon5DGPU::CreateComputeShader(LPCTSTR pFile, ID3D11ComputeShader **shader)
{
	// Open compiled shader file
	HANDLE fileHandle = CreateFile(pFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (fileHandle == INVALID_HANDLE_VALUE)
	{
		wprintf(L"Compute shader %s failed to load", pFile);
		exit(1);
	}

	DWORD fileSize = GetFileSize(fileHandle, NULL);
	if (fileSize == INVALID_FILE_SIZE)
	{
		wprintf(L"Compute shader %s failed to load", pFile);
		exit(1);
	}

	// Allocate memory to store the file data
	void *buffer = malloc(fileSize);
	if (buffer == NULL)
	{
		wprintf(L"Compute shader %s failed to load", pFile);
		exit(1);
	}

	// Read file from disk
	BOOL bhr;
	DWORD wmWritten;
	bhr = ReadFile(fileHandle, buffer, fileSize, &wmWritten, NULL);
	if (bhr == 0)
	{
		wprintf(L"Compute shader %s failed to load", pFile);
		exit(1);
	}

	// Close file
	CloseHandle(fileHandle);

	HRESULT hr = m_pDevice->CreateComputeShader(buffer, fileSize, nullptr, shader);
	free(buffer);

	if (FAILED(hr))
	{
		wprintf(L"Compute shader %s failed to load", pFile);
		exit(1);
	}
}
