#pragma once
/*
Copyright (c) 2010-2014, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    * Neither the name of Intel Corporation nor the names of its
      contributors may be used to endorse or promote products derived from
      this software without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

// A minimal task system based on a simplified version of the ispc task system
// We use the Concurrency runtime

#define NOMINMAX
#include <windows.h>

// Task system
// ========================================================================
#include <concrt.h>
using namespace Concurrency;

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <algorithm>

// Signature of  'task' functions
typedef void (*TaskFuncType)(void *data, int taskIndex, int taskCount);

// Small structure used to hold the data for each task
struct TaskInfo
{
	TaskFuncType func;
	void *data;
	int taskIndex, taskCount;
	event taskEvent;
};

///////////////////////////////////////////////////////////////////////////
// TaskGroupBase

#define LOG_TASK_QUEUE_CHUNK_SIZE 16
#define MAX_TASK_QUEUE_CHUNKS 8
#define TASK_QUEUE_CHUNK_SIZE (1 << LOG_TASK_QUEUE_CHUNK_SIZE)

/** The TaskGroup structure provides common functionality for "task
    groups"; a task group is the set of tasks launched from within a single
    ispc function.  When the function is ready to return, it waits for all
    of the tasks in its task group to finish before it actually returns.
 */
class TaskGroup
{
public:
	int AllocTaskInfo(int count);
	TaskInfo *GetTaskInfo(int index);

	void Launch(int count);
	void Sync();

	TaskGroup();
	~TaskGroup();

protected:
	int nextTaskInfoIndex;

private:
	/* We allocate blocks of TASK_QUEUE_CHUNK_SIZE TaskInfo structures as
	   needed by the calling function.  We hold up to MAX_TASK_QUEUE_CHUNKS
	   of these (and then exit at runtime if more than this many tasks are
	   launched.)
	 */
	TaskInfo *taskInfo[MAX_TASK_QUEUE_CHUNKS];
};

inline TaskGroup::TaskGroup()
{
	nextTaskInfoIndex = 0;

	for (int i = 0; i < MAX_TASK_QUEUE_CHUNKS; ++i)
	{
		taskInfo[i] = NULL;
	}
}

inline TaskGroup::~TaskGroup() {}

inline int TaskGroup::AllocTaskInfo(int count)
{
	int ret = nextTaskInfoIndex;
	nextTaskInfoIndex += count;
	return ret;
}

inline TaskInfo *TaskGroup::GetTaskInfo(int index)
{
	int chunk = (index >> LOG_TASK_QUEUE_CHUNK_SIZE);
	int offset = index & (TASK_QUEUE_CHUNK_SIZE - 1);

	if (chunk == MAX_TASK_QUEUE_CHUNKS)
	{
		fprintf(stderr, "A total of %d tasks have been launched from the "
		                "current function--the simple built-in task system can handle "
		                "no more. You can increase the values of TASK_QUEUE_CHUNK_SIZE "
		                "and LOG_TASK_QUEUE_CHUNK_SIZE to work around this limitation.  "
		                "Sorry!  Exiting.\n",
		        index);
		exit(1);
	}

	if (taskInfo[chunk] == NULL)
		taskInfo[chunk] = new TaskInfo[TASK_QUEUE_CHUNK_SIZE];
	return &taskInfo[chunk][offset];
}

///////////////////////////////////////////////////////////////////////////
// Concurrency Runtime

static void __cdecl RunTask(LPVOID param)
{
	TaskInfo *ti = (TaskInfo *)param;

	// Run the task.
	ti->func(ti->data, ti->taskIndex, ti->taskCount);

	// Signal the event that this task is done
	ti->taskEvent.set();
}

inline void TaskGroup::Launch(int count)
{
	for (int i = 0; i < count; ++i)
	{
		CurrentScheduler::ScheduleTask(RunTask, GetTaskInfo(i));
	}
}

inline void TaskGroup::Sync()
{
	for (int i = 0; i < nextTaskInfoIndex; ++i)
	{
		TaskInfo *ti = GetTaskInfo(i);
		ti->taskEvent.wait();
		ti->taskEvent.reset();
	}
}
